"""Plotting functions using [`matplotlib`](https://matplotlib.org/) library.

Warning:
   You need to have [`matplotlib`](https://matplotlib.org/) installed in order to use this module.
"""

from typing import Literal
import numpy as np

import matplotlib.pyplot as plt
from matplotlib import scale as mscale
from matplotlib import transforms as mtransforms
from matplotlib.ticker import NullLocator, FixedLocator
from matplotlib.ticker import ScalarFormatter, NullFormatter
from matplotlib.lines import Line2D
from matplotlib.axes import Axes

from acoustic_toolbox.bands import octave, third


class OctaveBandScale(mscale.ScaleBase):
    """Octave band scale."""

    name = "octave"

    def __init__(self, axis, **kwargs):
        mscale.ScaleBase.__init__(self, axis)

    def get_transform(self):
        return self.BandTransform()

    def set_default_locators_and_formatters(self, axis):
        axis.set_major_locator(FixedLocator(octave(16, 16000)))
        axis.set_major_formatter(ScalarFormatter())
        axis.set_minor_locator(NullLocator())
        axis.set_minor_formatter(NullFormatter())

    def limit_range_for_scale(self, vmin, vmax, minpos):
        return vmin, vmax

    class BandTransform(mtransforms.Transform):
        input_dims = 1
        output_dims = 1
        is_separable = True
        has_inverse = False

        def __init__(self):
            mtransforms.Transform.__init__(self)

        def transform_non_affine(self, a):
            return np.log10(a + 1)


mscale.register_scale(OctaveBandScale)


class ThirdBandScale(mscale.ScaleBase):
    """Third-octave band scale."""

    name = "third"

    def __init__(self, axis, **kwargs):
        mscale.ScaleBase.__init__(self, axis)

    def get_transform(self):
        return self.BandTransform()

    def set_default_locators_and_formatters(self, axis):
        axis.set_major_locator(NullLocator())
        axis.set_major_formatter(ScalarFormatter())
        axis.set_minor_locator(FixedLocator(third(12.5, 20000)))
        axis.set_minor_formatter(ScalarFormatter())

    def limit_range_for_scale(self, vmin, vmax, minpos):
        return vmin, vmax

    class BandTransform(mtransforms.Transform):
        input_dims = 1
        output_dims = 1
        is_separable = True
        has_inverse = False

        def __init__(self):
            mtransforms.Transform.__init__(self)

        def transform_non_affine(self, a):
            return np.log10(a + 1)


mscale.register_scale(ThirdBandScale)


def plot_octave(
    data: np.ndarray,
    octaves: np.ndarray,
    axes: Axes | None = None,
    kHz: bool = False,
    xlabel: str | None = None,
    ylabel: str | None = None,
    title: str | None = None,
    separator: str | None = None,
    *args,
    **kwargs,
) -> list[Line2D]:
    """Plot octave bands from `data` levels and `octaves` bands.

    Args:
      data: levels in a 1-D NumPy array.
      octaves: octaves in a 1-D NumPy array.
        Note that you can use [`acoustic_toolbox.bands.octave`][acoustic_toolbox.bands.octave]
        for this or manually enter all bands.
      axes: a `matplotlib.axes` object.
      kHz: if `True` it shows "1k" or "2k" instead of "1000" or "2000" as
      tick labels.
      xlabel: a `str` containing label for x axis.
      ylabel: a `str` containing label for y axis.
      title: a `str` containing title.
      separator: a `str` defining the decimal separator.  By default takes '.'
        or ',' values according to system settings (when separator is None).
      *args:
      **kwargs:

    Returns:
      A list of [`matplotlib.lines.Line2D`][matplotlib.lines.Line2D] objects.

    See Also:
      [`matplotlib.axes.Axes.plot`][matplotlib.axes.Axes.plot]
    """
    band_type = "octave"
    k_ticks = kHz
    return plot_bands(
        data,
        octaves,
        axes,
        band_type,
        k_ticks,
        xlabel,
        ylabel,
        title,
        separator,
        *args,
        **kwargs,
    )


def plot_third(
    data: np.ndarray,
    thirds: np.ndarray,
    axes: Axes | None = None,
    kHz: bool = False,
    xlabel: str | None = None,
    ylabel: str | None = None,
    title: str | None = None,
    separator: str | None = None,
    *args,
    **kwargs,
) -> list[Line2D]:
    """Plot third octave bands from `data` levels and `thirds` bands.

    Args:
      data: levels in an 1-D NumPy array.
      thirds: thirds in an 1-D NumPy array. Note that you can use
        [`acoustic_toolbox.bands.third`][acoustic_toolbox.bands.third] for this
        or manually enter all bands.
      axes: a `matplotlib.axes` object.
      kHz: if `True` it shows "1k" or "2.5k" instead of "1000" or "2500" as
        tick labels.
      xlabel: a `str` containing label for x axis (optional).
      ylabel: a `str` containing label for y axis (optional).
      title: a `str` containing title (optional).
      separator: a `str` defining the decimal separator. By default takes '.'
        or ',' values according to system settings (when separator is None).
      *args:
      **kwargs:

    Returns:
      A list of [`matplotlib.lines.Line2D`][matplotlib.lines.Line2D] objects.

    See Also:
      [`matplotlib.axes.Axes.plot`][matplotlib.axes.Axes.plot]
    """
    band_type = "third"
    k_ticks = kHz
    return plot_bands(
        data,
        thirds,
        axes,
        band_type,
        k_ticks,
        xlabel,
        ylabel,
        title,
        separator,
        *args,
        **kwargs,
    )


def plot_bands(
    data: np.ndarray,
    bands: np.ndarray,
    axes: Axes,
    band_type: Literal["octave", "third"],
    k_ticks: bool = False,
    xlabel: str | None = None,
    ylabel: str | None = None,
    title: str | None = None,
    separator: str | None = None,
    *args,
    **kwargs,
) -> list[Line2D]:
    """Plot bands from `data` levels and `bands`.

    Only use if you want to plot from arbitrary octave or third octave data.

    Args:
      data: levels in an 1-D NumPy array.
      bands: bands in an 1-D NumPy array.
      axes: `matplotlib.axes` object.
      band_type: `'octave'` or `'third'` are accepted values.
      k_ticks: if `True` it shows "1k" or "2.5k" instead of "1000" or "2500" as
        tick labels.
      xlabel: label for x axis.
      ylabel: label for y axis.
      title: title.
      separator: decimal separator. By default takes '.' or ',' values
        according to system settings (when separator is None).
      *args: additional arguments for `matplotlib.axes.plot`.
      **kwargs: additional arguments for `matplotlib.axes.plot`.

    Returns:
      A list of [`matplotlib.lines.Line2D`][matplotlib.lines.Line2D] objects.

    See Also:
      [`matplotlib.axes.Axes.plot`][matplotlib.axes.Axes.plot]
    """
    if axes is None:
        axes = plt.gca()
    factor = 0.1
    min_auto = bands[0] * (1 - factor)
    max_auto = bands[-1] * (1 + factor)
    axes.set_xlim([min_auto, max_auto])
    axes.set_xscale(band_type)

    # Set tick labels.
    ticklabels = _get_ticklabels(band_type, k_ticks, separator)
    if band_type == "third":
        third_ticks = True
    elif band_type == "octave":
        third_ticks = False
    axes.set_xticklabels(ticklabels, minor=third_ticks)

    # Set x and y labels and title.
    if xlabel is not None:
        axes.set_xlabel(xlabel)
    if ylabel is not None:
        axes.set_ylabel(ylabel)
    if title is not None:
        axes.set_title(title)
    return axes.plot(bands, data, *args, **kwargs)


TICKS_OCTAVE = [
    "16",
    "31.5",
    "63",
    "125",
    "250",
    "500",
    "1000",
    "2000",
    "4000",
    "8000",
    "16000",
]
"""Octave center frequencies as strings."""

TICKS_OCTAVE_KHZ = [
    "16",
    "31.5",
    "63",
    "125",
    "250",
    "500",
    "1k",
    "2k",
    "4k",
    "8k",
    "16k",
]
"""Octave center frequencies as strings. Uses kHz notation."""

TICKS_THIRD_OCTAVE = [
    "12.5",
    "16",
    "20",
    "25",
    "31.5",
    "40",
    "50",
    "63",
    "80",
    "100",
    "125",
    "160",
    "200",
    "250",
    "315",
    "400",
    "500",
    "630",
    "800",
    "1000",
    "1250",
    "1600",
    "2000",
    "2500",
    "3150",
    "4000",
    "5000",
    "6300",
    "8000",
    "10000",
    "12500",
    "16000",
    "20000",
]
"""Third-octave center frequencies as strings."""

TICKS_THIRD_OCTAVE_KHZ = [
    "12.5",
    "16",
    "20",
    "25",
    "31.5",
    "40",
    "50",
    "63",
    "80",
    "100",
    "125",
    "160",
    "200",
    "250",
    "315",
    "400",
    "500",
    "630",
    "800",
    "1000",
    "1250",
    "1600",
    "2000",
    "2500",
    "3150",
    "4000",
    "5000",
    "6300",
    "8000",
    "10000",
    "12500",
    "16000",
    "20000",
]
"""Third-octave center frequencies as strings. Uses kHz notation."""


def _get_ticklabels(band_type, kHz, separator):
    """Return a list with all labels for octave or third-octave band cases."""
    if separator is None:
        import locale

        separator = locale.localeconv()["decimal_point"]

    if band_type == "octave":
        if kHz is True:
            ticklabels = TICKS_OCTAVE_KHZ
        else:
            ticklabels = TICKS_OCTAVE
    else:
        if kHz is True:
            ticklabels = TICKS_THIRD_OCTAVE_KHZ
        else:
            ticklabels = TICKS_THIRD_OCTAVE

    ticklabels = _set_separator(ticklabels, separator)
    return ticklabels


def _set_separator(ticklabels, separator):
    """Set the decimal separator. Note that this is a 'private' function, so you can set the decimal separator directly in plotting functions."""
    if separator == ".":
        bands_sep = ticklabels
    else:
        bands_sep = []
        for item in ticklabels:
            decimal_number_format = item.replace(".", separator)
            bands_sep.append(decimal_number_format)
    return bands_sep



"""ISO 1996-2:2007

ISO 1996-2:2007 describes how sound pressure levels can be determined by direct measurement,
by extrapolation of measurement results by means of calculation, or exclusively by calculation,
intended as a basis for assessing environmental noise.

Reference:
    ISO 1996-2:2007: Description, measurement and assessment of environmental noise
"""

import numpy as np
import pandas as pd
from scipy.signal import welch
from scipy.stats import linregress
import matplotlib.pyplot as plt
from acoustic_toolbox.decibel import dbsum
from acoustic_toolbox.standards.iso_tr_25417_2007 import REFERENCE_PRESSURE
import weakref
from tabulate import tabulate

TONE_WITHIN_PAUSE_CRITERION_DB = 6.0
"""A tone may exist when the level of any line in the noise pause is 6 dB or more about...."""

TONE_BANDWIDTH_CRITERION_DB = 3.0
"""Bandwidth of the detected peak."""

TONE_LINES_CRITERION_DB = 6.0
"""All lines with levels within 6 dB of the maximum level are classified as tones."""

TONE_SEEK_CRITERION = 1.0
"""Tone seek criterion."""

REGRESSION_RANGE_FACTOR = 0.75
"""Range of regression is usually +/- 0.75 critical bandwidth."""

_WINDOW_CORRECTION = {
    "hann": -1.8,
}
"""Window correction factors for different window types."""


def window_correction(window):
    """Get correction factor to be applied to $L_{pt}$ due to window type.

    Args:
        window: Window type (e.g., 'hann').

    Returns:
        float: Correction factor in dB.

    Raises:
        ValueError: If window correction is not available for specified window.

    See Also:
        Tonality: Class that uses window correction in tonal analysis
    """
    try:
        return _WINDOW_CORRECTION[window]
    except KeyError:
        raise ValueError("Window correction is not available for specified window.")


def critical_band(frequency):
    """Bandwidth of critical band of frequency.

    Args:
        frequency: Center frequency of tone in Hz.

    Returns:
        tuple: A tuple containing:
            - float: Center frequency (minimum 50 Hz)
            - float: Lower band-edge frequency
            - float: Upper band-edge frequency
            - float: Bandwidth (100 Hz below 500 Hz, 20% of center frequency above)

    See Also:
        tonal_audibility: Function that uses critical band parameters
    """
    if isinstance(frequency, np.ndarray):
        center = frequency.copy()
        center[frequency < 50.0] = 50.0
    else:
        center = 50.0 if frequency < 50 else frequency

    bandwidth = (center > 500.0) * (center * 0.20) + (center <= 500.0) * 100.0

    upper = center + bandwidth / 2.0
    lower = center - bandwidth / 2.0

    return center, lower, upper, bandwidth


def tones_level(tone_levels):
    r"""Total sound pressure level of the tones in a critical band given the level of each of the tones.

    Returns:
        float: Total sound pressure level Lpt calculated as:
            $$
            L_{pt} = 10 \log_{10}{\sum 10^{L_{pti}/10}}
            $$

    Note:
        Implementation of equation C.1 from section C.2.3.1 of the standard.

    See Also:
        masking_noise_level: Function to calculate masking noise level
    """
    return dbsum(tone_levels)


def masking_noise_level(
    noise_lines, frequency_resolution, effective_analysis_bandwidth
):
    r"""Masking noise level $L_{pn}$.

    Args:
        noise_lines: Array of masking noise lines Ln in dB.
        frequency_resolution: Frequency resolution Δf in Hz.
        effective_analysis_bandwidth: Effective analysis bandwidth B in Hz.

    Returns:
        float: Masking noise level calculated as:
            $$
            L_{pn} = 10 \log_{10}{\sum 10^{L_n/10}} + 10 \log_{10}{\frac{\Delta f}{B}}
            $$

    Note:
        Implementation of equation C.11 from section C.4.4 of the standard.

    See Also:
        tones_level: Function to calculate total sound pressure level of tones
    """
    return dbsum(noise_lines) + 10.0 * np.log10(
        frequency_resolution / effective_analysis_bandwidth
    )


def masking_noise_lines(
    levels: pd.Series,
    line_classifier,
    center: float,
    bandwidth: float,
    regression_range_factor,
) -> tuple[np.ndarray, float, float]:
    """Determine masking noise level lines using regression line. Returns array of $L_n$.

    Args:
        levels: Levels as function of frequency
        line_classifier: Categorical indicating line types.
        center: Center frequency in Hz.
        bandwidth: Critical band bandwidth in Hz.
        regression_range_factor: Range factor for regression analysis.

    Returns:
        (ndarray): Array of masking noise lines Ln
        (float): Regression slope
        (float): Regression intercept
    """
    slicer = slice(
        center - bandwidth * regression_range_factor,
        center + bandwidth * regression_range_factor,
    )
    levels = levels[slicer]
    frequencies = levels.index
    regression_levels = levels[line_classifier == "noise"]
    slope, intercept = linregress(x=regression_levels.index, y=regression_levels)[0:2]
    levels_from_regression = slope * frequencies + intercept
    return levels_from_regression, slope, intercept


def tonal_audibility(tones_level, masking_noise_level, center) -> float:
    r"""Tonal audibility.

    Args:
        tones_level: Total sound pressure level of tones in critical band Lpt.
        masking_noise_level: Total sound pressure level of masking noise Lpn.
        center: Center frequency of critical band fc.

    Returns:
        Tonal audibility ΔLta calculated as:
            $$
            \Delta L_{ta} = L_{pt} - L_{pn} + 2 + \log_{10}{1 + \left(\frac{f_c}{502}\right)^{2.5}}
            $$


    Note:
        Implementation of equation C.3 from section C.2.4 of the standard.

    See Also:
        critical_band: Function to calculate critical band parameters
        tonal_adjustment: Function to calculate tonal adjustment
    """
    return (
        tones_level
        - masking_noise_level
        + 2.0
        + np.log10(1.0 + (center / 502.0) ** (2.5))
    )


def tonal_adjustment(tonal_audibility):
    """Calculate tonal adjustment Kt.

    Args:
        tonal_audibility: Tonal audibility $L_{ta}$ in dB.

    Returns:
        float: Adjustment Kt in dB:
            - 6.0 dB if ΔLta > 10 dB
            - 0.0 dB if ΔLta < 4 dB
            - ΔLta - 4 dB otherwise

    Note:
        Implementation of equations C.4-C.6 from section C.2.4 of the standard.

    See Also:
        tonal_audibility: Function to calculate tonal audibility
    """
    if tonal_audibility > 10.0:
        return 6.0
    elif tonal_audibility < 4.0:
        return 0.0
    else:
        return tonal_audibility - 4.0


class Tonality:
    """Perform assessment of audibility of tones in noise.

    Objective method for assessing the audibility of tones in noise.

    Args:
        signal: Time-domain signal samples.
        sample_frequency: Sample frequency in Hz.
        window: Window type for spectral analysis. Defaults to 'hann'.
        reference_pressure: Reference pressure. Defaults to REFERENCE_PRESSURE.
        tsc: Tone seeking criterion in dB. Defaults to TONE_SEEK_CRITERION.
        regression_range_factor: Regression range factor. Defaults to REGRESSION_RANGE_FACTOR.
        nbins: Number of frequency bins for FFT. Defaults to sample_frequency.
        force_tone_without_pause: Force tone detection without noise pause. Defaults to False.
        force_bandwidth_criterion: Force bandwidth criterion. Defaults to False.
    """

    def __init__(  # pylint: disable=too-many-instance-attributes
        self,
        signal,
        sample_frequency,
        window="hann",
        reference_pressure=REFERENCE_PRESSURE,
        tsc=TONE_SEEK_CRITERION,
        regression_range_factor=REGRESSION_RANGE_FACTOR,
        nbins=None,
        force_tone_without_pause=False,
        force_bandwidth_criterion=False,
    ):
        self.signal = signal
        """Samples in time-domain."""
        self.sample_frequency = sample_frequency
        """Sample frequency."""
        self.window = window
        """Window to be used."""
        self.reference_pressure = reference_pressure
        """Reference sound pressure."""
        self.tsc = tsc
        """Tone seeking criterium."""
        self.regression_range_factor = regression_range_factor
        """Regression range factor."""
        self.nbins = nbins
        """Amount of frequency nbins to use. See attribute `nperseg` of :func:`scipy.signal.welch`."""

        self._noise_pauses = list()
        """Private list of noise pauses that were determined or assigned."""
        self._spectrum = None
        """Power spectrum as function of frequency."""

        self.force_tone_without_pause = force_tone_without_pause
        self.force_bandwidth_criterion = force_bandwidth_criterion

    @property
    def noise_pauses(self):
        """Get determined noise pauses.

        Yields:
            NoisePause: Each noise pause found in the signal.
        """
        for noise_pause in self._noise_pauses:
            yield noise_pause

    @property
    def tones(self):
        """Get determined tones.

        Yields:
            Tone: Each tone found in noise pauses.
        """
        for noise_pause in self.noise_pauses:
            if noise_pause.tone is not None:
                yield noise_pause.tone

    @property
    def critical_bands(self):
        """Get critical bands.

        A critical band is determined for each detected tone.

        Yields:
            CriticalBand: Each critical band around detected tones.
        """
        for tone in self.tones:
            yield tone.critical_band

    @property
    def spectrum(self):
        """Get power spectrum of the input signal.

        Returns:
            pandas.Series: Power spectrum in dB re reference_pressure.
        """
        if self._spectrum is None:
            nbins = self.nbins
            if nbins is None:
                nbins = self.sample_frequency
            nbins //= 1  # Fix because of bug in welch with uneven nbins
            f, p = welch(
                self.signal,
                fs=self.sample_frequency,
                nperseg=nbins,
                window=self.window,
                detrend=False,
                scaling="spectrum",
            )
            self._spectrum = pd.Series(
                10.0 * np.log10(p / self.reference_pressure**2.0), index=f
            )
        return self._spectrum

    @property
    def frequency_resolution(self):
        """Frequency resolution.

        Returns:
            float: Frequency resolution Δf in Hz.
        """
        df = np.diff(np.array(self.spectrum.index)).mean()
        return df

    @property
    def effective_analysis_bandwidth(self):
        r"""Effective analysis bandwidth.

        In the case of the Hanning window
        $$
        B_{eff} = 1.5 \Delta f
        $$

        with $\Delta f$ the `frequency_resolution`.

        Returns:
            float: Effective analysis bandwidth in Hz.

        Raises:
            ValueError: If window type is not supported.

        Note:
            See section C.2.2 Note 1 of the standard.
        """
        if self.window == "hann":
            return 1.5 * self.frequency_resolution
        else:
            raise ValueError()

    def _set_noise_pauses(self, noise_pauses):
        """Manually set noise pauses. Expects iterable of tuples.

        Args:
            noise_pauses: Iterable of (start, end) tuples.

        Returns:
            self: For method chaining.
        """
        self._noise_pauses = [NoisePause(start, end) for start, end in noise_pauses]
        return self

    def determine_noise_pauses(self, end=None):
        """Find noise pauses in the spectrum.

        Uses noise_pause_seeker to find potential noise pauses.

        Args:
            end: Optional end index to limit search range.

        Returns:
            self: For method chaining.
        """
        self._set_noise_pauses(
            noise_pause_seeker(np.array(self.spectrum[:end]), self.tsc)
        )
        return self

    def _construct_line_classifier(self):
        """Set values of line classifier.

        Returns:
            self: For method chaining.
        """
        levels = self.spectrum

        categories = ["noise", "start", "end", "neither", "tone"]
        self.line_classifier = pd.Series(
            pd.Categorical(["noise"] * len(levels), categories=categories),
            index=levels.index,
        )

        # Add noise pauses
        for noise_pause in self.noise_pauses:
            # Mark noise pause start and end.
            self.line_classifier.iloc[noise_pause.start] = "start"
            self.line_classifier.iloc[noise_pause.end] = "end"
            # Mark all other lines within noise pause as neither tone nor noise.
            self.line_classifier.iloc[noise_pause.start + 1 : noise_pause.end] = (
                "neither"  # Half-open interval
            )

        # Add tone lines
        for tone in self.tones:
            self.line_classifier.iloc[tone._tone_lines] = "tone"

        return self

    def _determine_tones(self):
        """Analyze noise pauses for tones.

        Examines each noise pause for potential tones using determine_tone_lines.
        Creates Tone objects for any detected tones.

        Returns:
            self: For method chaining.
        """
        levels = self.spectrum

        # First we need to check for the tones.
        for noise_pause in self.noise_pauses:
            # Determine the indices of the tones in a noise pause
            tone_indices, bandwidth_for_tone_criterion = determine_tone_lines(
                levels,
                self.frequency_resolution,
                noise_pause.start,
                noise_pause.end,
                self.force_tone_without_pause,
                self.force_bandwidth_criterion,
            )
            # If we have indices, ...
            if np.any(tone_indices):
                # ...then we create a tone object.
                noise_pause.tone = create_tone(
                    levels,
                    tone_indices,
                    bandwidth_for_tone_criterion,
                    weakref.proxy(noise_pause),
                )
        return self

    def _determine_critical_bands(self):
        """Put a critical band around each of the determined tones.

        Returns:
            self: For method chaining.
        """
        for tone in self.tones:
            critical_band = self.critical_band_at(tone.center)
            tone.critical_band = critical_band
            critical_band.tone = weakref.proxy(tone)
        return self

    def analyse(self):
        """Analyse the noise pauses for tones and put critical bands around each of these tones.

        The tones are available via `tones` and the critical bands via `critical_bands`.
        Per frequency line results are available via `line_classifier`.

        Returns:
            self: For method chaining.
        """
        # Determine tones. Puts noise pause starts/ends in classier as well as tone lines
        # and lines that are neither tone nor noise.
        self._determine_tones()
        # Construct line classifier
        self._construct_line_classifier()
        # Determine critical bands.
        self._determine_critical_bands()
        return self

    def critical_band_at(self, frequency):
        """Put at a critical band at `frequency`.

        In order to use this function `line_classifier` needs to be available,
        which means `analyse` needs to be used first.

        Args:
            frequency: Center frequency in Hz.

        Returns:
            CriticalBand: Critical band object.

        Note:
            Requires line_classifier to be available (call analyse first).
        """
        return create_critical_band(
            self.spectrum,
            self.line_classifier,
            frequency,
            self.frequency_resolution,
            self.effective_analysis_bandwidth,
            self.regression_range_factor,
            self.window,
        )

    def plot_spectrum(self):
        """Plot power spectrum.

        Returns:
            matplotlib.figure.Figure: Figure with spectrum plot.
        """
        spectrum = self.spectrum
        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.plot(spectrum.index, spectrum)
        ax.set_xlabel("f in Hz")
        ax.set_ylabel("L in dB")
        return fig

    @property
    def dominant_tone(self):
        """Get most dominant tone.

        The most dominant tone is the one with highest tonal audibility $L_{ta}$.

        Returns:
            Tone: Most dominant tone, or None if no tones found.
        """
        try:
            return sorted(
                self.tones, key=lambda x: x.critical_band.tonal_audibility, reverse=True
            )[0]
        except IndexError:
            return None

    def plot_results(self, noise_pauses=False, tones=True, critical_bands=True):
        """Plot analysis results.

        Args:
            noise_pauses: Whether to show noise pauses. Defaults to False.
            tones: Whether to show tones. Defaults to True.
            critical_bands: Whether to show critical bands. Defaults to True.

        Returns:
            matplotlib.figure.Figure: Figure with results plot.
        """
        df = self.frequency_resolution
        levels = self.spectrum

        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.plot(levels.index, levels)
        ax.set_xlabel("$f$ in Hz")
        ax.set_ylabel("$L$ in dB")

        if noise_pauses:
            for pause in self.noise_pauses:
                ax.axvspan(pause.start * df, pause.end * df, color="green", alpha=0.05)

        if tones:
            for tone in self.tones:
                ax.axvline(tone.center, color="black", alpha=0.05)

        if critical_bands:
            for band in self.critical_bands:
                ax.axvspan(band.start, band.end, color="yellow", alpha=0.05)

        band = self.dominant_tone.critical_band
        ax.axvline(band.start, color="red", linewidth=0.1)
        ax.axvline(band.end, color="red", linewidth=0.1)

        # Limit xrange
        if noise_pauses:
            _items = list(self.noise_pauses)
        elif critical_bands:
            _items = list(self.critical_bands)
        ax.set_xlim(
            min(item.start for item in _items), max(item.end for item in _items)
        )
        return fig

    def overview(self):
        """Print overview of analysis results.

        Returns:
            str: Tabulated overview of results.

        Raises:
            ValueError: If no tones have been determined yet.
        """
        try:
            cb = self.dominant_tone.critical_band
        except AttributeError:
            raise ValueError(
                "Cannot show overview (yet). No tones have been determined."
            )

        tones = [
            ("Tone", "{:4.1f} Hz: {:4.1f} dB".format(tone.center, tone.tone_level))
            for tone in self.tones
        ]

        table = [
            ("Critical band", "{:4.1f} to {:4.1f} Hz".format(cb.start, cb.end)),
            (
                "Masking noise level $L_{pn}$",
                "{:4.1f} dB".format(cb.masking_noise_level),
            ),
            ("Tonal level $L_{pt}$", "{:4.1f} dB".format(cb.total_tone_level)),
            ("Dominant tone", "{:4.1f} Hz".format(cb.tone.center)),
            (
                "3 dB bandwidth of tone",
                "{:2.1f}% of {:4.1f}".format(
                    cb.tone.bandwidth_3db / cb.bandwidth * 100.0, cb.bandwidth
                ),
            ),
            ("Tonal audibility $L_{ta}$", "{:4.1f} dB".format(cb.tonal_audibility)),
            ("Adjustment $K_{t}$", "{:4.1f} dB".format(cb.adjustment)),
            ("Frequency resolution", "{:4.1f} Hz".format(self.frequency_resolution)),
            (
                "Effective analysis bandwidth",
                "{:4.1f} Hz".format(self.effective_analysis_bandwidth),
            ),
        ]
        table += tones
        return tabulate(table)

    def results_as_dataframe(self):
        """Get analysis results as pandas DataFrame.

        Returns:
            pandas.DataFrame: DataFrame containing results for each tone.
        """
        data = (
            (
                tone.center,
                tone.tone_level,
                tone.bandwidth_3db,
                tone.critical_band.start,
                tone.critical_band.end,
                tone.critical_band.bandwidth,
                tone.critical_band.regression_slope,
                tone.critical_band.regression_intercept,
                tone.critical_band.masking_noise_level,
                tone.critical_band.total_tone_level,
                tone.critical_band.tonal_audibility,
                tone.critical_band.adjustment,
            )
            for tone in self.tones
        )
        columns = [
            "center",
            "tone_level",
            "bandwidth_3db",
            "critical_band_start",
            "critical_band_end",
            "critical_band_bandwidth",
            "regression_slope",
            "regression_intercept",
            "masking_noise_level",
            "total_tone_level",
            "tonal_audibility",
            "adjustment",
        ]
        return pd.DataFrame(list(data), columns=columns)


class NoisePause:
    """Container for noise pause information.

    A noise pause is a section of the spectrum that may contain a tone.

    Args:
        start: Start index of noise pause.
        end: End index of noise pause.
        tone: Optional Tone object if a tone is found in this pause.
    """

    def __init__(self, start, end, tone=None):
        self.start = start
        self.end = end
        self.tone = tone

    def __str__(self):
        return "(start={},end={})".format(self.start, self.end)

    def __repr__(self):
        return "NoisePause{}".format(str(self))

    def __iter__(self):
        yield self.start
        yield self.stop

    def _repr_html_(self):
        """HTML representation for Jupyter notebooks."""
        table = [("Start", self.start), ("End", self.end)]
        return tabulate(table, tablefmt="html")


def create_tone(levels, tone_lines, bandwidth_for_tone_criterion, noise_pause):
    """Create an instance of Tone."""
    center = levels.iloc[tone_lines].idxmax()
    tone_level = tones_level(levels.iloc[tone_lines])
    return Tone(
        center, tone_lines, tone_level, noise_pause, bandwidth_for_tone_criterion
    )


class Tone:
    """Container for tone information.

    Args:
        center: Center frequency in Hz.
        tone_lines: Indices of spectral lines belonging to tone.
        tone_level: Total level of tone in dB.
        noise_pause: Parent NoisePause object.
        bandwidth_3db: -3 dB bandwidth in Hz.
        critical_band: Optional CriticalBand object.
    """

    def __init__(
        self,
        center,
        tone_lines,
        tone_level,
        noise_pause,
        bandwidth_3db,
        critical_band=None,
    ):
        self.center = center
        self._tone_lines = tone_lines
        self.tone_level = tone_level
        self.noise_pause = noise_pause
        self.bandwidth_3db = bandwidth_3db
        self.critical_band = critical_band

    def __str__(self):
        return "(center={:4.1f}, levels={:4.1f})".format(self.center, self.tone_level)

    def __repr__(self):
        return "Tone{}".format(str(self))

    def _repr_html_(self):
        """HTML representation for Jupyter notebooks."""
        table = [
            ("Center frequency", "{:4.1f} Hz".format(self.center)),
            ("Tone level", "{:4.1f} dB".format(self.tone_level)),
        ]
        return tabulate(table, tablefmt="html")


def create_critical_band(
    levels,
    line_classifier,
    frequency,
    frequency_resolution,
    effective_analysis_bandwidth,
    regression_range_factor,
    window,
    tone=None,
):
    """Create an instance of CriticalBand."""
    center, start, end, bandwidth = critical_band(frequency)

    # Masking noise lines
    noise_lines, regression_slope, regression_intercept = masking_noise_lines(
        levels, line_classifier, center, bandwidth, regression_range_factor
    )
    # Masking noise level
    noise_level = masking_noise_level(
        noise_lines, frequency_resolution, effective_analysis_bandwidth
    )
    # Total tone level
    tone_lines = levels[line_classifier == "tone"][start:end]
    tone_level = tones_level(tone_lines) - window_correction(window)
    # Tonal audibility
    audibility = tonal_audibility(tone_level, noise_level, center)
    # Adjustment Kt
    adjustment = tonal_adjustment(audibility)

    return CriticalBand(
        center,
        start,
        end,
        bandwidth,
        regression_range_factor,
        regression_slope,
        regression_intercept,
        noise_level,
        tone_level,
        audibility,
        adjustment,
        tone,
    )


class CriticalBand:
    """Container for critical band information.

    A critical band is a frequency band around a tone used for masking analysis.

    Args:
        center: Center frequency in Hz.
        start: Lower band-edge frequency in Hz.
        end: Upper band-edge frequency in Hz.
        bandwidth: Bandwidth in Hz.
        regression_range_factor: Factor for regression range.
        regression_slope: Slope from linear regression.
        regression_intercept: Intercept from linear regression.
        noise_level: Masking noise level in dB.
        tone_level: Total tone level in dB.
        audibility: Tonal audibility in dB.
        adjustment: Tonal adjustment in dB.
        tone: Optional Tone object.
    """

    def __init__(  # pylint: disable=too-many-instance-attributes
        self,
        center,
        start,
        end,
        bandwidth,
        regression_range_factor,
        regression_slope,
        regression_intercept,
        noise_level,
        tone_level,
        audibility,
        adjustment,
        tone=None,
    ):
        self.center = center
        """Center frequency of the critical band."""
        self.start = start
        """Lower band-edge frequency of the critical band."""
        self.end = end
        """Upper band-edge frequency of the critical band."""
        self.bandwidth = bandwidth
        """Bandwidth of the critical band."""
        self.regression_range_factor = regression_range_factor
        """Range of regression factor. See also :attr:`REGRESSION_RANGE_FACTOR`."""
        self.regression_slope = regression_slope
        """Linear regression slope."""
        self.regression_intercept = regression_intercept
        """Linear regression intercept."""
        self.masking_noise_level = noise_level
        """Masking noise level $L_{pn}$."""
        self.total_tone_level = tone_level
        """Total tone level $L_{pt}$."""
        self.tonal_audibility = audibility
        """Tonal audibility $L_{ta}$."""
        self.adjustment = adjustment
        """Adjustment $K_{t}$."""
        self.tone = tone

    def __str__(self):
        return "(center={:4.1f}, bandwidth={:4.1f}, tonal_audibility={:4.1f}, adjustment={:4.1f}".format(
            self.center, self.bandwidth, self.tonal_audibility, self.adjustment
        )

    def __repr__(self):
        return "CriticalBand{}".format(str(self))

    def _repr_html_(self):
        """HTML representation for Jupyter notebooks."""
        table = [
            ("Center frequency", "{:4.1f} Hz".format(self.center)),
            ("Start frequency", "{:4.1f} Hz".format(self.start)),
            ("End frequency", "{:4.1f} Hz".format(self.end)),
            ("Bandwidth", "{:4.1f} Hz".format(self.bandwidth)),
            ("Regression factor", "{:4.1f}".format(self.regression_range_factor)),
            ("Regression slope", "{:4.1f}".format(self.regression_slope)),
            ("Regression intercept", "{:4.1f}".format(self.regression_intercept)),
            ("Masking noise level", "{:4.1f} dB".format(self.masking_noise_level)),
            ("Total tone level", "{:4.1f} dB".format(self.total_tone_level)),
            ("Tonal audibility $L_{ta}$", "{:4.1f} dB".format(self.tonal_audibility)),
            ("Adjustment $K_{t}$", "{:4.1f} dB".format(self.adjustment)),
        ]

        return tabulate(table, tablefmt="html")


# ----------Noise pauses----------------------------


def _search_noise_pauses(levels, tsc):
    """Search for noise pauses in a level sequence.

    Args:
        levels: Array of level values.
        tsc: Tone seeking criterion in dB.

    Returns:
        list: List of tuples containing (start_index, end_index) for each noise pause.
    """
    pauses = list()
    possible_start = None
    for i in range(2, len(levels) - 2):
        if (levels[i] - levels[i - 1]) >= tsc and (levels[i - 1] - levels[i - 2]) < tsc:
            possible_start = i
        if (levels[i] - levels[i + 1]) >= tsc and (levels[i + 1] - levels[i + 2]) < tsc:
            if possible_start:
                pauses.append((possible_start, i))
                possible_start = None
    return pauses


def noise_pause_seeker(levels, tsc):
    """Given the levels of a spectrum and a tone seeking criterium
    this top level function seeks possible noise pauses.

    Args:
        levels: Spectral levels in dB.
        tsc: Tone seeking criterion in dB.

    Returns:
        list: List of tuples containing (start_index, end_index) for each noise pause,
            sorted and filtered to avoid overlapping intervals.

    Note:
    Possible start and end indices of noise pauses are determined using `possible_noise_pauses.

    Then, only those that correspond to the smallest intervals that do not overlap other intervals are kept.
    """
    n = len(levels)
    forward_pauses = _search_noise_pauses(levels, tsc)
    backward_pauses = _search_noise_pauses(levels[::-1], tsc)
    backward_pauses = [
        (n - 1 - start, n - 1 - end) for end, start in reversed(backward_pauses)
    ]
    possible_pauses = sorted(list(set(forward_pauses) & set(backward_pauses)))
    return possible_pauses


# ------------------- Tone seeking--------------------


def determine_tone_lines(
    levels,
    df,
    start,
    end,
    force_tone_without_pause=False,
    force_bandwidth_criterion=False,
):
    """Determine tone lines in a noise pause.

    Args:
        levels: Series with levels as function of frequency.
        df: Frequency resolution in Hz.
        start: Index of noise pause start.
        end: Index of noise pause end.
        force_tone_without_pause: Force tone detection without pause. Defaults to False.
        force_bandwidth_criterion: Force bandwidth criterion. Defaults to False.

    Returns:
        tuple: A tuple containing:
            - ndarray: Indices of tone lines
            - float: -3 dB bandwidth in Hz
    """
    # Noise pause range object
    npr = slice(start, end + 1)

    # Return values
    tone_indices = np.array([])
    bandwidth_for_tone_criterion = None

    # Levels but with integeres as indices instead of frequencies.
    # Benefit over np.array is that the index is maintained when the object is sliced.
    levels_int = levels.reset_index(drop=True)

    # If any of the lines is six 6 dB above. See section C.4.3.
    if (
        np.any(
            (
                levels.iloc[npr]
                >= TONE_WITHIN_PAUSE_CRITERION_DB + levels.iloc[start - 1]
            )
            & (
                levels.iloc[npr]
                >= TONE_WITHIN_PAUSE_CRITERION_DB + levels.iloc[end + 1]
            )
        )
        or force_tone_without_pause
    ):
        # Indices of values that are within -3 dB point.
        indices_3db = (
            (levels.iloc[npr] >= levels.iloc[npr].max() - TONE_BANDWIDTH_CRITERION_DB)
            .to_numpy()
            .nonzero()[0]
        )
        # -3 dB bandwidth
        bandwidth_for_tone_criterion = (indices_3db.max() - indices_3db.min()) * df
        # Frequency of tone.
        tone_center_frequency = levels.iloc[npr].idxmax()
        # tone_center_index = levels.reset_index(drop=True).iloc[npr].idxmax()
        # Critical band
        _, _, _, critical_band_bandwidth = critical_band(tone_center_frequency)

        # Fullfill bandwidth criterion? See section C.4.3
        if (
            bandwidth_for_tone_criterion < 0.10 * critical_band_bandwidth
        ) or force_bandwidth_criterion:
            # All values within 6 decibel are designated as tones.
            tone_indices = (
                levels_int.iloc[npr][
                    levels_int.iloc[npr]
                    >= levels_int.iloc[npr].max() - TONE_LINES_CRITERION_DB
                ]
            ).index.values

    return tone_indices, bandwidth_for_tone_criterion



import numpy as np
from scipy.signal import sawtooth
from acoustic_toolbox.cepstrum import complex_cepstrum, inverse_complex_cepstrum
from numpy.testing import assert_array_almost_equal


class TestCepstrum:
    """Test :mod:`acoustic_toolbox.cepstrum`"""

    def test_complex_cepstrum(self):
        """The period of a periodic harmonic will show up as a peak in a
        complex cepstrum.
        """
        duration = 5.0
        fs = 8000.0
        samples = int(fs * duration)
        t = np.arange(samples) / fs
        fundamental = 100.0
        signal = sawtooth(2.0 * np.pi * fundamental * t)
        ceps, _ = complex_cepstrum(signal)
        assert fundamental == 1.0 / t[ceps.argmax()]

    def test_inverse_complex_cepstrum(self):
        """Applying the complex cepstrum and then the inverse complex cepstrum
        should result in the original sequence.
        """
        x = np.arange(10)
        ceps, ndelay = complex_cepstrum(x)
        y = inverse_complex_cepstrum(ceps, ndelay)
        assert_array_almost_equal(x, y)



"""IEC 61672-1:2013

This module implements IEC 61672-1:2013 which provides electroacoustical performance specifications
for three kinds of sound measuring instruments:

1. Time-weighting sound level meters that measure exponential-time-weighted, frequency-weighted sound levels
2. Integrating-averaging sound level meters that measure time-averaged, frequency-weighted sound levels
3. Integrating sound level meters that measure frequency-weighted sound exposure levels

The module provides functions for:
- Frequency weighting (A, C, Z)
- Time weighting (Fast, Slow)
- Decibel level calculations

It uses the pyoctaveband package for accurate time and frequency weighting filters.

Reference:
    IEC 61672-1:2013: http://webstore.iec.ch/webstore/webstore.nsf/artnum/048669
    pyoctaveband package: https://pypi.org/project/pyoctaveband/
"""

import io
import os
import pkgutil
import numpy as np
import pandas as pd
from typing import Tuple, Dict
from numpy.typing import NDArray
from scipy.signal import sosfilt, sosfreqz
from .iso_tr_25417_2007 import REFERENCE_PRESSURE

from pyoctaveband import time_weighting, WeightingFilter

WEIGHTING_DATA = pd.read_csv(
    io.BytesIO(
        pkgutil.get_data(
            "acoustic_toolbox", os.path.join("data", "iec_61672_1_2013.csv")
        )
    ),
    sep=",",
    index_col=0,
)
"""DataFrame with indices, nominal frequencies and weighting values."""

NOMINAL_THIRD_OCTAVE_CENTER_FREQUENCIES: NDArray[np.float64] = np.array(
    WEIGHTING_DATA.nominal
)
"""Nominal 1/3-octave frequencies. See table 3."""

NOMINAL_OCTAVE_CENTER_FREQUENCIES: NDArray[np.float64] = np.array(
    WEIGHTING_DATA.nominal
)[2::3]
"""Nominal 1/1-octave frequencies. Based on table 3."""

REFERENCE_FREQUENCY: float = 1000.0
"""Reference frequency. See table 3."""

EXACT_THIRD_OCTAVE_CENTER_FREQUENCIES: NDArray[np.float64] = (
    REFERENCE_FREQUENCY * 10.0 ** (0.01 * (np.arange(10, 44) - 30))
)
"""Exact third-octave center frequencies. See table 3."""

WEIGHTING_A: NDArray[np.float64] = np.array(WEIGHTING_DATA.A)
"""Frequency weighting A. See table 3."""

WEIGHTING_C: NDArray[np.float64] = np.array(WEIGHTING_DATA.C)
"""Frequency weighting C. See table 3."""

WEIGHTING_Z: NDArray[np.float64] = np.array(WEIGHTING_DATA.Z)
"""Frequency weighting Z. See table 3."""

WEIGHTING_VALUES: Dict[str, NDArray[np.float64]] = {
    "A": WEIGHTING_A,
    "C": WEIGHTING_C,
    "Z": WEIGHTING_Z,
}
"""Dictionary with weighting values 'A', 'C' and 'Z' weighting."""


FAST: float = 0.125
"""FAST time-constant."""

SLOW: float = 1.000
"""SLOW time-constant."""


def time_averaged_level(
    signal: NDArray[np.float64],
    sample_frequency: float,
    integration_time: float,
    reference: float = REFERENCE_PRESSURE,
) -> Tuple[NDArray[np.float64], NDArray[np.float64]]:
    """Calculate time-averaged level.

    Args:
        signal: Dynamic pressure.
        sample_frequency: Sample frequency in Hz.
        integration_time: Integration time in seconds.
        reference: Reference for decibels. Defaults to REFERENCE_PRESSURE.

    Returns:
        Tuple containing:
            - Time points in seconds
            - Time-averaged levels in dB
    """
    signal = np.asarray(signal)
    integration_time = np.asarray(integration_time)
    sample_frequency = np.asarray(sample_frequency)
    signal_samples = signal.shape[-1]

    step = integration_time * sample_frequency
    n_steps = int(signal_samples / step)

    # Calculate end indices for each chunk (exclusive)
    target_boundaries = np.arange(1, n_steps + 1) * step
    end_indices = np.floor(target_boundaries - 1e-9).astype(int) + 1

    # Calculate start indices for reduceat
    start_indices = np.concatenate(([0], end_indices[:-1]))

    # Truncate signal to the end of the last full chunk
    limit = end_indices[-1]
    sq_signal_truncated = signal[..., :limit]**2.0

    # Sum squared values in each chunk
    sums = np.add.reduceat(sq_signal_truncated, start_indices, axis=-1)

    # Calculate chunk lengths (number of samples per chunk)
    lengths = np.diff(np.concatenate(([0], end_indices)))

    # Compute mean squared pressure
    means = sums / lengths

    levels = 10.0 * np.log10(means / reference**2.0)
    times = np.arange(levels.shape[-1]) * integration_time
    return times, levels


def time_weighted_level(
    signal: NDArray[np.float64],
    sample_frequency: float,
    time_mode: str,
    integration_time: float = None,
    reference: float = REFERENCE_PRESSURE,
) -> Tuple[NDArray[np.float64], NDArray[np.float64]]:
    """Calculate time-weighted levels at a given integration timestep.

    Args:
        signal: raw signal (can be multi-dimensional).
        sample_frequency: Sample frequency in Hz.
        time_mode: Time weighting mode, either "fast" or "slow"
        integration_time: timestep in seconds. defaults to 125ms for "fast" mode and 1s for "slow" mode.
        reference: Reference for decibels. Defaults to REFERENCE_PRESSURE.

    Returns:
        Tuple containing:
            - Time points in seconds
            - Time-weighted levels in dB
    """
    if time_mode.lower() not in ["fast", "slow"]:
        raise ValueError("time_mode must be either 'fast' or 'slow'.")

    if integration_time is None:
        integration_time = (FAST if time_mode.lower() == "fast" else SLOW)

    signal_samples = signal.shape[-1]

    # get time-weighted squared signal
    tw_sq_signal = time_weighting(signal, sample_frequency, mode=time_mode.lower())

    step = integration_time * sample_frequency
    n_steps = int(signal_samples / step)

    indices = np.arange(1, n_steps + 1) * step
    # select the last sample before each integration time step
    # we use - 1e-9 so that :
    # - when step is an integer (e.g. 4000 * 0.125 = 500), we get the correct index (499 instead of 500)
    # - when step is not an integer (e.g. 44100 * 0.125 = 5512.5), we get the correct index (5512 instead of 5512.5)
    indices = np.floor(indices - 1e-9).astype(int)

    tw_sq_values = tw_sq_signal[..., indices]

    levels = 10.0 * np.log10(tw_sq_values / reference**2.0)
    times = np.arange(levels.shape[-1]) * integration_time

    return times, levels




def frequency_weighting(
    signal: NDArray[np.float64],
    sample_frequency: float,
    weighting: str = 'A',
    zero_phase: bool = False
) -> NDArray[np.float64]:
    """Apply frequency weighting to a signal.

    Args:
        signal: Input signal (raw pressure/voltage).
        sample_frequency: Sample rate.
        weighting: 'A', 'C' or 'Z'.
        zero_phase: Prevent phase shift.
            If True, processing is done in the frequency domain (FFT) to ensure
            zero phase shift while preserving the exact IEC 61672-1 magnitude response.
            (Note: Standard forward-backward filtering would square the magnitude response).

    Returns:
        Frequency-weighted signal.
    """
    if weighting == "Z":
        return signal

    wf = WeightingFilter(fs=sample_frequency, curve=weighting)

    if zero_phase:
        # Frequency domain filtering (zero phase)
        # Get frequency response
        w, h = sosfreqz(wf.sos, worN=signal.shape[-1], fs=sample_frequency)
        # FFT of the signal
        signal_fft = np.fft.rfft(signal, axis=-1)
        # Interpolate filter response to match FFT bins
        h_interp = np.interp(
            np.fft.rfftfreq(signal.shape[-1], d=1/sample_frequency),
            w,
            np.abs(h)
        )
        # Apply filter in frequency domain
        weighted_signal_fft = signal_fft * h_interp
        # Inverse FFT to get time-domain signal
        weighted_signal = np.fft.irfft(weighted_signal_fft, n=signal.shape[-1], axis=-1)
    else:
        # Time domain filtering (causal)
        weighted_signal = sosfilt(wf.sos, signal, axis=-1)

    return weighted_signal



"""Signal class.

This module contains the `Signal` class, which is a container for time-domain signals.

Attributes:
    fs: Sample frequency.
    samples: Amount of samples in signal.
    channels: Amount of channels in signal.
    duration: Duration of signal in seconds.
    values: Values of signal as instance of [`np.ndarray`][numpy.ndarray].

Methods:
    calibrate_to: Calibrate signal to value `decibel`.
    calibrate_with: Calibrate signal with other signal.
    decimate: Decimate signal by integer `factor`.
    resample: Resample signal.
    upsample: Upsample signal with integer factor.
    gain: Apply gain of `decibel` decibels.
    pick: Get signal from start time to stop time.
    times: Time vector.
    energy: Signal energy.
    power: Signal power.
    ms: Mean value squared of signal.
    rms: Root mean squared of signal.
    weigh: Apply frequency-weighting.
    instantaneous_frequency: Instantaneous frequency.
    instantaneous_phase: Instantaneous phase.
    detrend: Detrend signal.
    unwrap: Unwrap signal.
    complex_cepstrum: Complex cepstrum.
    real_cepstrum: Real cepstrum.
    power_spectrum: Power spectrum.
    angle_spectrum: Angle spectrum.
    phase_spectrum: Phase spectrum.
    peak: Peak value.
    peak_level: Peak level.
    min: Minimum value.
    max: Maximum value.
    max_level: Maximum level.
    sound_exposure: Sound exposure.
    sound_exposure_level: Sound exposure level.
    plot_complex_cepstrum: Plot complex cepstrum.
    plot_real_cepstrum: Plot real cepstrum.
    plot_power_spectrum: Plot power spectrum.
    plot_angle_spectrum: Plot angle spectrum.
    plot_phase_spectrum: Plot phase spectrum.
    spectrogram: Spectrogram.
    plot_spectrogram: Plot spectrogram.
    fast_levels: Fast time-weighted level.
    slow_levels: Slow time-weighted level.
    leq_levels: Time-weighted levels.
    levels: Levels (deprecated).
    leq: Leq.
    plot_levels: Plot levels.
    bandpass: Bandpass filter.
    bandstop: Bandstop filter.
    highpass: Highpass filter.
    lowpass: Lowpass filter.
    octavepass: Octavepass filter.
    third_octaves: Third octaves.
    fractional_octaves: Fractional octaves.
    plot_octaves: Plot octaves.
    plot_third_octaves: Plot third octaves.
    plot_fractional_octaves: Plot fractional octaves.
    plot: Plot signal.
    normalize: Normalize signal.
    to_wav: Write signal to WAV file.
    from_wav: Read signal from WAV file.

See Also:
    - [`acoustic_toolbox.signal`](signal.md)
"""

from __future__ import annotations
import itertools
import warnings
from matplotlib.axes import Axes
import matplotlib.pyplot as plt
import numpy as np
from scipy.io import wavfile
from scipy.signal import (
    detrend,
    lfilter,
    bilinear,
    spectrogram,
    filtfilt,
    resample,
    fftconvolve,
)
from acoustic_toolbox import signal
from acoustic_toolbox.cepstrum import complex_cepstrum, real_cepstrum
from acoustic_toolbox import standards
from acoustic_toolbox.signal import Frequencies

from acoustic_toolbox.standards.iso_tr_25417_2007 import REFERENCE_PRESSURE
from acoustic_toolbox.standards.iec_61672_1_2013 import (
    frequency_weighting,
    time_averaged_level,
    time_weighted_level,
)
from acoustic_toolbox.standards.iec_61672_1_2013 import (
    NOMINAL_OCTAVE_CENTER_FREQUENCIES,
    NOMINAL_THIRD_OCTAVE_CENTER_FREQUENCIES,
)


class Signal(np.ndarray):
    """A signal consisting of samples (array) and a sample frequency (float).

    Attributes:
        fs: Sample frequency.
        samples: Amount of samples in signal.
        channels: Amount of channels in signal.
        duration: Duration of signal in seconds.
        values: Values of signal as instance of [`np.ndarray`][numpy.ndarray].
    """

    def __new__(cls, data, fs):
        """Create a new signal.

        Args:
            data: Signal values.
            fs: Sample frequency.
        """
        obj = np.asarray(data).view(cls)
        obj.fs = fs
        return obj

    def __array_prepare__(self, array, context=None):
        try:
            # Best guess: the context here is the existing signal, which may have multiple channels.
            # This is checking for (1) is it two channels and (2) are the sample frequencies the same.
            # Issues: only works for two channels, not n channels (and just seems inelegant).
            a = context[1][0]
            b = context[1][1]
        except IndexError:
            return array

        if hasattr(a, "fs") and hasattr(b, "fs"):
            if a.fs == b.fs:
                return array
            else:
                raise ValueError("Sample frequencies do not match.")
        else:
            return array

    def __array_wrap__(self, out_arr, context=None, return_scalar: bool = False):
        if np.lib.NumpyVersion(np.__version__) >= "2.0.0":
            return np.ndarray.__array_wrap__(self, out_arr, context, return_scalar)
        else:
            return np.ndarray.__array_wrap__(self, out_arr, context)

    def __array_finalize__(self, obj):
        # see InfoArray.__array_finalize__ for comments
        if obj is None:
            return

        self.fs = getattr(obj, "fs", None)

    def __reduce__(self):
        # Get the parent's __reduce__ tuple
        pickled_state = super(Signal, self).__reduce__()
        # Create our own tuple to pass to __setstate__
        new_state = pickled_state[2] + (self.fs,)
        # Return a tuple that replaces the parent's __setstate__ tuple with our own
        return (pickled_state[0], pickled_state[1], new_state)

    def __setstate__(self, state):
        self.fs = state[-1]  # Set the info attribute
        # Call the parent's __setstate__ with the other tuple elements.
        super(Signal, self).__setstate__(state[0:-1])

    def __repr__(self):
        return "Signal({})".format(str(self))

    def _construct(self, x):
        """Construct signal like x."""
        return Signal(x, self.fs)

    @property
    def samples(self):
        """Amount of samples in signal."""
        return self.shape[-1]

    @property
    def channels(self):
        """Amount of channels."""
        if self.ndim > 1:
            return self.shape[-2]
        else:
            return 1

    @property
    def duration(self) -> float:
        """Duration of signal in seconds."""
        return float(self.samples / self.fs)

    @property
    def values(self) -> np.ndarray:
        """Return the values of this signal as an instance of [`np.ndarray`][numpy.ndarray]."""
        return np.array(self)

    def calibrate_to(self, decibel, inplace: bool = False) -> Signal:
        """Calibrate signal to value `decibel`.

        Tip:
            Values of `decibel` are broadcasted. To set a value per channel, use `decibel[...,None]`.

        Args:
            decibel: Value to calibrate to.
            inplace: Whether to perform inplace or not.

        Returns:
            Calibrated signal.
        """
        decibel = decibel * np.ones(self.shape)
        gain = decibel - self.leq()[..., None]
        return self.gain(gain, inplace=inplace)

    def calibrate_with(self, other: Signal, decibel, inplace: bool = False) -> Signal:
        """Calibrate signal with other signal.

        Args:
          other: Other signal/array.
          decibel: Signal level of `other`.
          inplace: Whether to perform inplace or not.

        Returns:
          Calibrated signal.
        """
        if not isinstance(other, Signal):
            other = Signal(other, self.fs)
        gain = decibel - other.leq()
        return self.gain(gain, inplace=inplace)

    def decimate(self, factor, zero_phase=False, ftype="iir", order=None) -> Signal:
        """Decimate signal by integer `factor`. Before downsampling a low-pass filter is applied.

        Args:
            factor: Downsampling factor.
            zero_phase: Prevent phase shift by filtering with ``filtfilt`` instead of ``lfilter``.
            ftype: Filter type.
            order: Filter order.

        Returns:
            Decimated signal.

        .. seealso:: :func:`scipy.signal.decimate`
        .. seealso:: :meth:`resample`: Decimated signal.

        """
        return Signal(
            signal.decimate(
                x=self, q=factor, n=order, ftype=ftype, zero_phase=zero_phase
            ),
            self.fs / factor,
        )

    def resample(self, nsamples, times=None, axis=-1, window=None) -> Signal:
        """Resample signal.

        Tip:
            You might want to low-pass filter this signal before resampling.

        Args:
          samples: New amount of samples.
          times: Times corresponding to samples.
          axis: Axis.
          window: Window.

        Returns:
            Resampled signal.
        See Also:
            - [`scipy.signal.resample`][scipy.signal.resample]
            - [`decimate`][acoustic_toolbox._signal.Signal.decimate]

        """
        return Signal(
            resample(self, nsamples, times, axis, window),
            nsamples / self.samples * self.fs,
        )

    def upsample(self, factor: int, axis: int = -1) -> Signal:
        """Upsample signal with integer factor.

        Args:
          factor: Upsample factor.
          axis: Axis.

        See Also:
            - [`resample`][acoustic_toolbox._signal.Signal.resample]
        """
        return self.resample(int(self.samples * factor), axis=axis)

    def gain(self, decibel, inplace: bool = False) -> Signal:
        """Apply gain of `decibel` decibels.

        Args:
          decibel: Decibels
          inplace: In place

        Returns:
          Amplified signal.

        """
        factor = 10.0 ** (decibel / 20.0)
        if inplace:
            self *= factor
            return self
        else:
            return self * factor

    def pick(self, start: float = 0.0, stop: float | None = None) -> Signal:
        """Get signal from start time to stop time.

        Args:
          start: Start time.
          stop: End time.

        Returns:
          Selected part of the signal.

        """
        if start is not None:
            start = int(np.floor(start * self.fs))
        if stop is not None:
            stop = int(np.floor(stop * self.fs))
        return self[..., start:stop]

    def times(self) -> np.ndarray:
        """Time vector.

        Returns:
          A vector with a timestamp for each sample.

        """
        return np.arange(0, self.samples) / self.fs

    def energy(self) -> np.ndarray:
        r"""Signal energy.

        Returns:
            Total energy per channel.

            $$
            E = \sum_{n=0}^{N-1} |x_n|^2
            $$

        """
        return float((self * self).sum())

    def power(self):
        r"""Signal power.

        $$
        P = \frac{1}{N} \sum_{n=0}^{N-1} |x_n|^2
        $$
        """
        return self.energy() / len(self)

    def ms(self):
        """Mean value squared of signal.

        See Also:
            [`acoustic_toolbox.signal.ms`][acoustic_toolbox._signal.Signal.ms]
        """
        return signal.ms(self)

    def rms(self):
        """Root mean squared of signal.

        See Also:
            [`acoustic_toolbox.signal.rms`][acoustic_toolbox._signal.Signal.rms]
        """
        return signal.rms(self)
        # return np.sqrt(self.power())

    def weigh(self, weighting: str = "A", zero_phase: bool = False) -> Signal:
        """Apply frequency-weighting. By default 'A'-weighting is applied.

        Note:
            By default the weighting filter is applied in the time domain using
            [`scipy.signal.sosfilt`][scipy.signal.sosfilt] causing a frequency-dependent delay.

            In case a delay is undesired, the filter can be applied in the frequency domain using
            [`scipy.signal.fft`][scipy.signal.fft] by setting `zero_phase=True`.

        Args:
            weighting: Frequency-weighting filter to apply.
                       Valid options are 'A', 'C' and 'Z'. Default weighting is 'A'.
            zero_phase: Prevent phase shift by filtering with ``fft`` instead of ``sosfilt``.

        Returns:
            Weighted signal.
        """
        return self._construct(
            frequency_weighting(
                self, self.fs, weighting=weighting, zero_phase=zero_phase
            )
        )

    def correlate(self, other: Signal | None = None, mode: str = "full"):
        """Correlate signal with `other` signal. In case `other==None` this
        method returns the autocorrelation.

        Args:
          other: Other signal.
          mode: Mode.

        Raises:
            ValueError: If sample frequencies are not the same.
            ValueError: If not supported for multichannel signals.

        See Also:
            - [`np.correlate`][numpy.correlate]
            - [`scipy.signal.fftconvolve`][scipy.signal.fftconvolve]
        """
        if other is None:
            other = self
        if self.fs != other.fs:
            raise ValueError("Cannot correlate. Sample frequencies are not the same.")
        if self.channels > 1 or other.channels > 1:
            raise ValueError(
                "Cannot correlate. Not supported for multichannel signals."
            )
        return self._construct(fftconvolve(self, other[::-1], mode=mode))

    def amplitude_envelope(self) -> Signal:
        """Amplitude envelope.

        Returns:
            Amplitude envelope.

        See Also:
            - [`acoustic_toolbox.signal.amplitude_envelope`][acoustic_toolbox._signal.Signal.amplitude_envelope]

        """
        return self._construct(signal.amplitude_envelope(self, self.fs))

    def instantaneous_frequency(self) -> Signal:
        """Instantaneous frequency.

        Returns:
            Instantaneous frequency of signal

        See Also:
            - [`acoustic_toolbox.signal.instantaneous_frequency`][acoustic_toolbox._signal.Signal.instantaneous_frequency]


        """
        return self._construct(signal.instantaneous_frequency(self, self.fs))

    def instantaneous_phase(self) -> Signal:
        """Instantaneous phase.

        Returns:
            Instantaneous phase of signal

        See Also:
            - [`acoustic_toolbox.signal.instantaneous_phase`][acoustic_toolbox._signal.Signal.instantaneous_phase]
        """
        return self._construct(signal.instantaneous_phase(self, self.fs))

    def detrend(self, **kwargs) -> Signal:
        """Detrend signal.

        Args:
            **kwargs:

        Returns:
            Detrended signal.

        See Also:
            - [`scipy.signal.detrend`][scipy.signal.detrend]

        """
        return self._construct(detrend(self, **kwargs))

    def unwrap(self) -> Signal:
        """Unwrap signal in case the signal represents wrapped phase.

        Returns:
            Unwrapped signal.

        See Also:
            - [`np.unwrap`][numpy.unwrap]
        """
        return self._construct(np.unwrap(self))

    def complex_cepstrum(
        self, N: int | None = None
    ) -> tuple[np.ndarray, np.ndarray, int]:
        """Complex cepstrum.

        Args:
            N: Amount of bins.

        Returns:
            Quefrency
            Complex cepstrum
            Delay in amount of samples.

        See Also:
            - [`acoustic_toolbox.cepstrum.complex_cepstrum`][acoustic_toolbox._signal.Signal.complex_cepstrum]

        """
        if N is not None:
            times = np.linspace(0.0, self.duration, N, endpoint=False)
        else:
            times = self.times()
        cepstrum, ndelay = complex_cepstrum(self, n=N)
        return times, cepstrum, ndelay

    def real_cepstrum(self, N=None):
        """Real cepstrum.

        Args:
          N: Amount of bins.

        Returns:
          Quefrency
          Real cepstrum.

        See Also:
            - [`acoustic_toolbox.cepstrum.real_cepstrum`][acoustic_toolbox._signal.Signal.real_cepstrum]

        """
        if N is not None:
            times = np.linspace(0.0, self.duration, N, endpoint=False)
        else:
            times = self.times()
        return times, real_cepstrum(self, n=N)

    def power_spectrum(self, N: int | None = None):
        """Power spectrum.

        Args:
            N: Amount of bins.

        See Also:
            - [`acoustic_toolbox.signal.power_spectrum`][acoustic_toolbox._signal.Signal.power_spectrum]
        """
        return signal.power_spectrum(self, self.fs, N=N)

    def angle_spectrum(self, N: int | None = None):
        """Phase angle spectrum. Wrapped.

        Args:
          N: amount of bins.

        See Also:
            - [`acoustic_toolbox.signal.angle_spectrum`][acoustic_toolbox._signal.Signal.angle_spectrum]
            - [`acoustic_toolbox.signal.phase_spectrum`][acoustic_toolbox._signal.Signal.phase_spectrum]
            - [`phase_spectrum`][acoustic_toolbox._signal.Signal.phase_spectrum]
        """
        return signal.angle_spectrum(self, self.fs, N=N)

    def phase_spectrum(self, N=None):
        """Phase spectrum. Unwrapped.

        Args:
          N: Amount of bins.

        See Also:
            - [`acoustic_toolbox.signal.phase_spectrum`][acoustic_toolbox._signal.Signal.phase_spectrum]
            - [`acoustic_toolbox.signal.angle_spectrum`][acoustic_toolbox._signal.Signal.angle_spectrum]
            - [`angle_spectrum`][acoustic_toolbox._signal.Signal.angle_spectrum]
        """
        return signal.phase_spectrum(self, self.fs, N=N)

    def peak(self, axis=-1):
        """Peak sound pressure.

        Args:
          axis: Axis.

        See Also:
            - [`acoustic_toolbox.standards.iso_tr_25417_2007.peak_sound_pressure`][acoustic_toolbox.standards.iso_tr_25417_2007.peak_sound_pressure]
        """
        return standards.iso_tr_25417_2007.peak_sound_pressure(self, axis=axis)

    def peak_level(self, axis=-1):
        """Peak sound pressure level.

        Args:
          axis: Axis.

        See Also:
            - [`acoustic_toolbox.standards.iso_tr_25417_2007.peak_sound_pressure_level`][acoustic_toolbox.standards.iso_tr_25417_2007.peak_sound_pressure_level]
        """
        return standards.iso_tr_25417_2007.peak_sound_pressure_level(self, axis=axis)

    def min(self, axis=-1):
        """Return the minimum along a given axis.

        Args:
          axis: Axis.

        See Also:
            - Refer to [`np.amin`][numpy.amin] for full documentation.
        """
        return np.ndarray.min(self, axis=axis)

    def max(self, axis=-1):
        """Return the maximum along a given axis.

        Args:
          axis: Axis.

        See Also:
            - Refer to [`np.amax`][numpy.amax] for full documentation.
        """
        return np.ndarray.max(self, axis=axis)

    def max_level(self, axis=-1):
        """Maximum sound pressure level.

        Args:
          axis: Axis.

        See Also:
            - [`acoustic_toolbox.standards.iso_tr_25417_2007.max_sound_pressure_level`][acoustic_toolbox.standards.iso_tr_25417_2007.max_sound_pressure_level]
        """
        return standards.iso_tr_25417_2007.max_sound_pressure_level(self, axis=axis)

    def sound_exposure(self, axis=-1):
        """Sound exposure.

        Args:
          axis: Axis.

        See Also:
            - [`acoustic_toolbox.standards.iso_tr_25417_2007.sound_exposure`][acoustic_toolbox.standards.iso_tr_25417_2007.sound_exposure]
        """
        return standards.iso_tr_25417_2007.sound_exposure(self, self.fs, axis=axis)

    def sound_exposure_level(self, axis=-1):
        """Sound exposure level.

        Args:
          axis: Axis.

        See Also:
            - [`acoustic_toolbox.standards.iso_tr_25417_2007.sound_exposure_level`][acoustic_toolbox.standards.iso_tr_25417_2007.sound_exposure_level]
        """
        return standards.iso_tr_25417_2007.sound_exposure_level(
            self, self.fs, axis=axis
        )

    def plot_complex_cepstrum(self, N: int | None = None, **kwargs):
        """Plot complex cepstrum of signal.

        Args:
          N: Amount of bins.
          **kwargs:

        **Valid kwargs:**

        Other Args:
            xscale: X-axis scale.
            yscale: Y-axis scale.
            xlim: X-axis limits.
            ylim: Y-axis limits.
            frequency (bool): Boolean indicating whether the x-axis should show time in seconds or quefrency
            xlabel_frequency: Label in case frequency is shown.
        """
        params = {
            "xscale": "linear",
            "yscale": "linear",
            "xlabel": "$t$ in s",
            "ylabel": "$C$",
            "title": "Complex cepstrum",
            "frequency": False,
            "xlabel_frequency": "$f$ in Hz",
        }
        params.update(kwargs)

        t, ceps, _ = self.complex_cepstrum(N=N)
        if params["frequency"]:
            t = 1.0 / t
            params["xlabel"] = params["xlabel_frequency"]
            t = t[::-1]
            ceps = ceps[::-1]
        return _base_plot(t, ceps, params)

    def plot_real_cepstrum(self, N: int | None = None, **kwargs):
        """Plot real cepstrum of signal.

        Args:
          N: Amount of bins.
          **kwargs:

        **Valid kwargs:**

        Other Args:
            xscale: X-axis scale.
            yscale: Y-axis scale.
            xlim: X-axis limits.
            ylim: Y-axis limits.
            frequency (bool): Boolean indicating whether the x-axis should show time in seconds or quefrency
            xlabel_frequency: Label in case frequency is shown.
        """
        params = {
            "xscale": "linear",
            "yscale": "linear",
            "xlabel": "$t$ in s",
            "ylabel": "$C$",
            "title": "Real cepstrum",
            "frequency": False,
            "xlabel_frequency": "$f$ in Hz",
        }
        params.update(kwargs)

        t, ceps = self.real_cepstrum(N=N)
        if params["frequency"]:
            t = 1.0 / t
            params["xlabel"] = params["xlabel_frequency"]
            t = t[::-1]
            ceps = ceps[::-1]
        return _base_plot(t, ceps, params)

    def plot_power_spectrum(self, N=None, **kwargs):  # filename=None, scale='log'):
        """Plot spectrum of signal.

        Args:
          N: Amount of bins.
          **kwargs:

        **Valid kwargs:**

        Other Args:
            xscale: X-axis scale.
            yscale: Y-axis scale.
            xlim: X-axis limits.
            ylim: Y-axis limits.
            reference: Reference power

        See Also:
            - [`acoustic_toolbox.signal.power_spectrum`][acoustic_toolbox._signal.Signal.power_spectrum]
        """
        params = {
            "xscale": "log",
            "yscale": "linear",
            "xlabel": "$f$ in Hz",
            "ylabel": "$L_{p}$ in dB",
            "title": "SPL",
            "reference": REFERENCE_PRESSURE**2.0,
        }
        params.update(kwargs)

        f, o = self.power_spectrum(N=N)
        return _base_plot(f, 10.0 * np.log10(o / params["reference"]), params)

    def plot_angle_spectrum(self, N: int | None = None, **kwargs):
        """Plot phase angle spectrum of signal. Wrapped.

        Args:
          N: Amount of bins.
          **kwargs:

        **Valid kwargs:**

        Other Args:
            xscale: X-axis scale.
            yscale: Y-axis scale.
            xlim: X-axis limits.
            ylim: Y-axis limits.
            reference: Reference power
        """
        params = {
            "xscale": "linear",
            "yscale": "linear",
            "xlabel": "$f$ in Hz",
            "ylabel": r"$\angle \phi$",
            "title": "Phase response (wrapped)",
        }
        params.update(kwargs)
        f, o = self.angle_spectrum(N=N)
        return _base_plot(f, o, params)

    def plot_phase_spectrum(self, N: int | None = None, **kwargs):
        """Plot phase spectrum of signal. Unwrapped.

        Args:
          N: Amount of bins.
          **kwargs:

        **Valid kwargs:**

        Other Args:
            xscale: X-axis scale.
            yscale: Y-axis scale.
            xlim: X-axis limits.
            ylim: Y-axis limits.
            reference: Reference power
        """
        params = {
            "xscale": "linear",
            "yscale": "linear",
            "xlabel": "$f$ in Hz",
            "ylabel": r"$\angle \phi$",
            "title": "Phase response (unwrapped)",
        }
        params.update(kwargs)
        f, o = self.phase_spectrum(N=N)
        return _base_plot(f, o, params)

    def spectrogram(self, **kwargs):
        """Spectrogram of signal.

        Returns:
          Time
          Frequency
          Power

        See Also:
            See [`scipy.signal.spectrogram`][scipy.signal.spectrogram]. Some of the default values have been changed.
            The generated spectrogram consists by default of complex values.

        """
        params = {
            "nfft": 4096,
            "noverlap": 128,
            "mode": "complex",
        }
        params.update(kwargs)

        t, s, P = spectrogram(self, fs=self.fs, **params)

        return t, s, P

    def plot_spectrogram(self, **kwargs):
        """Plot spectrogram of the signal.

        Note:
            This method only works for a single channel.

        **Valid kwargs:**

        Other Args:
            xlim: X-axis limits.
            ylim: Y-axis limits.
            clim: Color limits.
            NFFT: Amount of FFT bins.
            noverlap: Amount of overlap between FFT bins.
            title: Title.
            xlabel: X-axis label.
            ylabel: Y-axis label.
            clabel: Color label.
            colorbar: Whether to show the colorbar.
        """
        # TODO: use `spectrogram`.
        params = {
            "xlim": None,
            "ylim": None,
            "clim": None,
            "NFFT": 4096,
            "noverlap": 128,
            "title": "Spectrogram",
            "xlabel": "$t$ in s",
            "ylabel": "$f$ in Hz",
            "clabel": "SPL in dB",
            "colorbar": True,
        }
        params.update(kwargs)

        if self.channels > 1:
            raise ValueError(
                "Cannot plot spectrogram of multichannel signal. Please select a single channel."
            )

        # Check if an axes object is passed in. Otherwise, create one.
        ax0 = params.get("ax", plt.figure().add_subplot(111))
        ax0.set_title(params["title"])

        data = np.squeeze(self)
        try:
            _, _, _, im = ax0.specgram(
                data,
                Fs=self.fs,
                noverlap=params["noverlap"],
                NFFT=params["NFFT"],
                mode="magnitude",
                scale_by_freq=False,
            )
        except AttributeError:
            raise NotImplementedError(
                "Your version of matplotlib is incompatible due to lack of support of the mode keyword argument to matplotlib.mlab.specgram."
            )

        if params["colorbar"]:
            cb = ax0.get_figure().colorbar(mappable=im)
            cb.set_label(params["clabel"])

        ax0.set_xlim(params["xlim"])
        ax0.set_ylim(params["ylim"])
        im.set_clim(params["clim"])

        ax0.set_xlabel(params["xlabel"])
        ax0.set_ylabel(params["ylabel"])

        return ax0

    def fast_levels(self, integration_time: float = 0.125):
        """Calculate the FAST time weighted level as every `integration_time` seconds.

        Args:
            integration_time: timestep for the output. Default value is 0.125 second (FAST) but can set to other values if desired.

        Returns:
            sound pressure level as function of time.

        See Also:
            - [`acoustic_toolbox.standards.iec_61672_1_2013.time_averaged_sound_level`][acoustic_toolbox.standards.iec_61672_1_2013.time_averaged_sound_level]
            - [`acoustic_toolbox.standards.iec_61672_1_2013.time_weighted_sound_level`][acoustic_toolbox.standards.iec_61672_1_2013.time_weighted_sound_level]

        """
        return standards.iec_61672_1_2013.time_weighted_level(
            self.values, self.fs, time_mode="fast", integration_time=integration_time
        )

    def slow_levels(self, integration_time: float = 1.0):
        """Calculate the SLOW time weighted level as every `integration_time` seconds.

        Args:
            integration_time: Averaging time constant. Default value is 1.0 second.
        Returns:
            sound pressure level as function of time.

        See Also:
            - [`acoustic_toolbox.standards.iec_61672_1_2013.time_averaged_sound_level`][acoustic_toolbox.standards.iec_61672_1_2013.time_averaged_sound_level]
            - [`acoustic_toolbox.standards.iec_61672_1_2013.time_weighted_sound_level`][acoustic_toolbox.standards.iec_61672_1_2013.time_weighted_sound_level]

        """
        return standards.iec_61672_1_2013.time_averaged_level(
            self.values, self.fs, time_mode="slow", integration_time=integration_time
        )

    def levels(self, time: float = 0.125, method: str = "average"):
        """Calculate sound pressure level as function of time.

        Args:
            time: Averaging time or integration time constant. Default value is 0.125 corresponding to FAST.
            method: Use time `average` or time `weighting`.

        Returns:
            sound pressure level as function of time.

        See Also:
            - [`acoustic_toolbox.standards.iec_61672_1_2013.time_averaged_level`][acoustic_toolbox.standards.iec_61672_1_2013.time_averaged_level]
            - [`acoustic_toolbox.standards.iec_61672_1_2013.time_weighted_level`][acoustic_toolbox.standards.iec_61672_1_2013.time_weighted_level]

        """
        warnings.warn(
            'Signal.levels is deprecated. Use Signal.fast_levels, Signal.slow_levels (for "weighting") or Signal.leq_levels (for "average") instead.',
            DeprecationWarning,
            stacklevel=2,
        )
        if method == "average":
            return time_averaged_level(self.values, self.fs, time)
        elif method == "weighting":
            if time == 0.125:
                time_mode = "fast"
            elif time == 1.0:
                time_mode = "slow"
            else:
                raise ValueError(
                    "Invalid time for weighting. Use 0.125 (FAST) or 1.0 (SLOW)."
                )
            return time_weighted_level(
                self.values, self.fs, time_mode=time_mode, integration_time=time
            )
        else:
            raise ValueError("Invalid method")

    def leq(self):
        """Equivalent level. Single-value number.

        See Also:
            - [`acoustic_toolbox.standards.iso_tr_25417_2007.equivalent_sound_pressure_level`][acoustic_toolbox.standards.iso_tr_25417_2007.equivalent_sound_pressure_level]
        """
        return standards.iso_tr_25417_2007.equivalent_sound_pressure_level(self.values)

    def plot_levels(self, **kwargs):
        """Plot sound pressure level as function of time.

        See Also:
            - [`levels`][acoustic_toolbox._signal.Signal.levels]
        """
        params = {
            "xscale": "linear",
            "yscale": "linear",
            "xlabel": "$t$ in s",
            "ylabel": "$L_{p,F}$ in dB",
            "title": "SPL",
            "time": 0.125,
            "method": "average",
            "labels": None,
        }
        params.update(kwargs)
        t, L = self.levels(params["time"], params["method"])
        L_masked = np.ma.masked_where(np.isinf(L), L)
        return _base_plot(t, L_masked, params)

    # def octave(self, frequency, fraction=1):
    # """Determine fractional-octave `fraction` at `frequency`.

    # .. seealso:: :func:`acoustic_toolbox.signal.fractional_octaves`

    # """
    # return acoustic_toolbox.signal.fractional_octaves(self, self.fs, frequency,
    # frequency, fraction, False)[1]

    def bandpass(self, lowcut, highcut, order=8, zero_phase=False) -> Signal:
        """Filter signal with band-pass filter.

        Args:
            lowcut: Lower cornerfrequency.
            highcut: Upper cornerfrequency.
            order: Filter order. (Default value = 8)
            zero_phase: Prevent phase error by filtering in both directions (filtfilt). (Default value = False)

        Returns:
            class:`Signal`.

        See Also:
            - [`acoustic_toolbox.signal.bandpass`][acoustic_toolbox.signal.bandpass]

        """
        return type(self)(
            signal.bandpass(
                self, lowcut, highcut, self.fs, order=order, zero_phase=zero_phase
            ),
            self.fs,
        )

    def bandstop(
        self, lowcut, highcut, order: int = 8, zero_phase: bool = False
    ) -> Signal:
        """Filter signal with band-stop filter.

        Args:
            lowcut: Lower cornerfrequency.
            highcut: Upper cornerfrequency.
            order: Filter order.
            zero_phase: Prevent phase error by filtering in both directions (filtfilt).

        Returns:
            Band-pass filtered signal

        See Also:
            - [`acoustic_toolbox.signal.bandstop`][acoustic_toolbox.signal.bandstop]

        """
        return type(self)(
            signal.bandstop(
                self, lowcut, highcut, self.fs, order=order, zero_phase=zero_phase
            ),
            self.fs,
        )

    def highpass(self, cutoff, order: int = 4, zero_phase: bool = False) -> Signal:
        """Filter signal with high-pass filter.

        Args:
            cutoff: Cornerfrequency.
            order: Filter order.
            zero_phase: Prevent phase error by filtering in both directions (filtfilt).

        Returns:
            High-pass filtered signal

        See Also:
            - [`acoustic_toolbox.signal.highpass`][acoustic_toolbox.signal.highpass]

        """
        return type(self)(
            signal.highpass(self, cutoff, self.fs, order=order, zero_phase=zero_phase),
            self.fs,
        )

    def lowpass(self, cutoff, order: int = 4, zero_phase: bool = False) -> Signal:
        """Filter signal with low-pass filter.

        Args:
            cutoff: Cornerfrequency.
            order: Filter order.
            zero_phase: Prevent phase error by filtering in both directions (filtfilt).

        Returns:
            Low-pass filtered signal

        See Also:
            - [`acoustic_toolbox.signal.lowpass`][acoustic_toolbox.signal.lowpass]

        """
        return type(self)(
            signal.lowpass(self, cutoff, self.fs, order=order, zero_phase=zero_phase),
            self.fs,
        )

    def octavepass(
        self, center, fraction, order: int = 8, zero_phase: bool = False
    ) -> Signal:
        """Filter signal with fractional-octave band-pass filter.

        Args:
            center: Center frequency. Any value in the band will suffice.
            fraction: Band designator.
            order: Filter order.
            zero_phase: Prevent phase error by filtering in both directions (filtfilt).

        Returns:
            Band-pass filtered signal

        See Also:
            - [`acoustic_toolbox.signal.octavepass`][acoustic_toolbox.signal.octavepass]

        """
        return type(self)(
            signal.octavepass(
                self,
                center,
                self.fs,
                fraction=fraction,
                order=order,
                zero_phase=zero_phase,
            ),
            self.fs,
        )

    def bandpass_frequencies(
        self,
        frequencies: "Frequencies",
        order: int = 8,
        purge: bool = True,
        zero_phase: bool = False,
    ) -> tuple["Frequencies", Signal]:
        """Apply bandpass filters for frequencies.

        Args:
            frequencies: Instance of :class:`acoustic_toolbox.signal.Frequencies`
            order: Filter order.
            purge: Discard bands of which the upper corner frequency is above the Nyquist frequency.
            zero_phase: Prevent phase error by filtering in both directions (filtfilt).

        Returns:
            Frequencies
            Band-pass filtered signal.

        See Also:
            - [`acoustic_toolbox.signal.bandpass_frequencies`][acoustic_toolbox.signal.bandpass_frequencies]

        """
        frequencies, filtered = signal.bandpass_frequencies(
            self, self.fs, frequencies, order, purge, zero_phase=zero_phase
        )
        return frequencies, type(self)(filtered, self.fs)

    def octaves(
        self,
        frequencies: "Frequencies" | np.ndarray = NOMINAL_OCTAVE_CENTER_FREQUENCIES,
        order: int = 8,
        purge: bool = True,
        zero_phase: bool = False,
    ):
        """Apply 1/1-octaves bandpass filters.

        Args:
            frequencies: Band-pass filter frequencies.
            order: Filter order.
            purge: Discard bands of which the upper corner frequency is above the Nyquist frequency.
            zero_phase: Prevent phase error by filtering in both directions (filtfilt).

        Returns:
            Frequencies
            Band-pass filtered signal.

        See Also:
            - [`acoustic_toolbox.signal.bandpass_octaves`][acoustic_toolbox.signal.bandpass_octaves]
        """
        frequencies, octaves = signal.bandpass_octaves(
            self, self.fs, frequencies, order, purge, zero_phase=zero_phase
        )
        return frequencies, type(self)(octaves, self.fs)

    def third_octaves(
        self,
        frequencies: "Frequencies" = NOMINAL_THIRD_OCTAVE_CENTER_FREQUENCIES,
        order: int = 8,
        purge: bool = True,
        zero_phase: bool = False,
    ):
        """Apply 1/3-octaves bandpass filters.

        Args:
            frequencies: Band-pass filter frequencies.
            order: Filter order.
            purge: Discard bands of which the upper corner frequency is above the Nyquist frequency.
            zero_phase: Prevent phase error by filtering in both directions (filtfilt).

        Returns:
          Frequencies and band-pass filtered signal.

        See Also:
            - [`acoustic_toolbox.signal.bandpass_third_octaves`][acoustic_toolbox.signal.bandpass_third_octaves]

        """
        frequencies, octaves = signal.bandpass_third_octaves(
            self, self.fs, frequencies, order, purge, zero_phase=zero_phase
        )
        return frequencies, type(self)(octaves, self.fs)

    def fractional_octaves(
        self,
        frequencies: "Frequencies" | None = None,
        fraction: int = 1,
        order: int = 8,
        purge: bool = True,
        zero_phase: bool = False,
    ):
        """Apply 1/N-octaves bandpass filters.

        Args:
            frequencies: Band-pass filter frequencies.
            fraction: Default band-designator of fractional-octaves.
            order: Filter order.
            purge: Discard bands of which the upper corner frequency is above the Nyquist frequency.
            zero_phase: Prevent phase error by filtering in both directions (filtfilt).

        Returns:
          Frequencies and band-pass filtered signal.

        See Also:
            - [`acoustic_toolbox.signal.bandpass_fractional_octaves`][acoustic_toolbox.signal.bandpass_fractional_octaves]

        """
        if frequencies is None:
            frequencies = signal.OctaveBand(
                fstart=NOMINAL_THIRD_OCTAVE_CENTER_FREQUENCIES[0],
                fstop=self.fs / 2.0,
                fraction=fraction,
            )
        frequencies, octaves = signal.bandpass_fractional_octaves(
            self, self.fs, frequencies, fraction, order, purge, zero_phase=zero_phase
        )
        return frequencies, type(self)(octaves, self.fs)

    def plot_octaves(self, **kwargs):
        """Plot octaves.

        See Also:
            - [`octaves`][acoustic_toolbox._signal.Signal.octaves]
        """
        params = {
            "xscale": "log",
            "yscale": "linear",
            "xlabel": "$f$ in Hz",
            "ylabel": "$L_{p}$ in dB",
            "title": "1/1-Octaves SPL",
        }
        params.update(kwargs)
        f, o = self.octaves()
        print(len(f.center), len(o.leq()))
        return _base_plot(f.center, o.leq().T, params)

    def plot_third_octaves(self, **kwargs):
        """Plot 1/3-octaves.

        See Also:
            - [`third_octaves`][acoustic_toolbox._signal.Signal.third_octaves]
        """
        params = {
            "xscale": "log",
            "yscale": "linear",
            "xlabel": "$f$ in Hz",
            "ylabel": "$L_{p}$ in dB",
            "title": "1/3-Octaves SPL",
        }
        params.update(kwargs)
        f, o = self.third_octaves()
        return _base_plot(f.center, o.leq().T, params)

    def plot_fractional_octaves(
        self,
        frequencies=None,
        fraction=1,
        order=8,
        purge=True,
        zero_phase=False,
        **kwargs,
    ):
        """Plot fractional octaves."""
        title = "1/{}-Octaves SPL".format(fraction)

        params = {
            "xscale": "log",
            "yscale": "linear",
            "xlabel": "$f$ in Hz",
            "ylabel": "$L_p$ in dB",
            "title": title,
        }
        params.update(kwargs)
        f, o = self.fractional_octaves(
            frequencies=frequencies,
            fraction=fraction,
            order=order,
            purge=purge,
            zero_phase=zero_phase,
        )
        return _base_plot(f.center, o.leq().T, params)

    def plot(self, **kwargs):
        """Plot signal as function of time. By default the entire signal is plotted.

        **Valid kwargs:**

        Other Args:
            filename: Name of file.
            start: First sample index.
            stop: Last sample index.
        """
        params = {
            "xscale": "linear",
            "yscale": "linear",
            "xlabel": "$t$ in s",
            "ylabel": "$x$ in -",
            "title": "Signal",
        }
        params.update(kwargs)
        return _base_plot(self.times(), self, params)

    # def plot_scalo(self, filename=None):
    # """
    # Plot scalogram
    # """
    # from scipy.signal import ricker, cwt

    # wavelet = ricker
    # widths = np.logspace(-1, 3.5, 10)
    # x = cwt(self, wavelet, widths)

    # interpolation = 'nearest'

    # from matplotlib.ticker import LinearLocator, AutoLocator, MaxNLocator
    # majorLocator = LinearLocator()
    # majorLocator = MaxNLocator()

    # fig = plt.figure()
    # ax = fig.add_subplot(111)
    # ax.set_title('Scaleogram')
    ##ax.set_xticks(np.arange(0, x.shape[1])*self.fs)
    ##ax.xaxis.set_major_locator(majorLocator)

    ##ax.imshow(10.0 * np.log10(x**2.0), interpolation=interpolation, aspect='auto', origin='lower')#, extent=[0, 1, 0, len(x)])
    # ax.pcolormesh(np.arange(0.0, x.shape[1])/self.fs, widths, 10.0*np.log(x**2.0))
    # if filename:
    # fig.savefig(filename)
    # else:
    # return fig

    # def plot_scaleogram(self, filename):
    # """
    # Plot scaleogram
    # """
    # import pywt

    # wavelet = 'dmey'
    # level = pywt.dwt_max_level(len(self), pywt.Wavelet(wavelet))
    # print level
    # level = 20
    # order = 'freq'
    # interpolation = 'nearest'

    # wp = pywt.WaveletPacket(self, wavelet, 'sym', maxlevel=level)
    # nodes = wp.get_level(level, order=order)
    # labels = [n.path for n in nodes]
    # values = np.abs(np.array([n.data for n in nodes], 'd'))

    # fig = plt.figure()
    # ax = fig.add_subplot(111)
    # ax.set_title('Scaleogram')
    # ax.imshow(values, interpolation=interpolation, aspect='auto', origin='lower', extent=[0, 1, 0, len(values)])
    ##ax.set_yticks(np.arange(0.5, len(labels) + 0.5))
    ##ax.set_yticklabels(labels)

    # fig.savefig(filename)

    def normalize(self, gap=6.0, inplace=False):
        """Normalize signal.

        The parameter `gap` can be understood as using `gap` decibels fewer for the dynamic range.

        By default a 6 decibel gap is used.

        Args:
          gap: Gap between maximum value and ceiling in decibel.
          inplace: Normalize signal in place.
        """
        factor = np.abs(self).max() * 10.0 ** (gap / 20.0)
        if inplace:
            self /= factor[..., None]
            return self
        else:
            return self / factor[..., None]

    def to_wav(self, filename, depth=16):
        """Save signal as WAV file.

        By default, this function saves a normalized 16-bit version of the signal with at least 6 dB range till clipping occurs.

        Args:
            filename: Name of file to save to.
            depth: If given, convert to integer with specified depth. Else, try to store using the original data type.

        """
        data = self
        dtype = data.dtype if not depth else "int" + str(depth)
        if depth:
            data = (data * 2 ** (depth - 1) - 1).astype(dtype)
        wavfile.write(filename, int(self.fs), data.T)
        # wavfile.write(filename, int(self.fs), self._data/np.abs(self._data).max() *  0.5)
        # wavfile.write(filename, int(self.fs), np.int16(self._data/(np.abs(self._data).max()) * 32767) )

    @classmethod
    def from_wav(cls, filename, normalize=True) -> Signal:
        """Create an instance of `Signal` from a WAV file.

        Args:
            filename: Filename of WAV file.
            normalize: Whether to normalize the signal.

        Returns:
            Signal
        """
        fs, data = wavfile.read(filename)
        data = data.astype(np.float32, copy=False).T
        if normalize:
            data /= np.max(np.abs(data))
        return cls(data, fs=fs)


_PLOTTING_PARAMS = {
    "title": None,
    "xlabel": None,
    "ylabel": None,
    "xscale": "linear",
    "yscale": "linear",
    "xlim": (None, None),
    "ylim": (None, None),
    "labels": None,
    "linestyles": ["-", "-.", "--", ":"],
}


def _get_plotting_params():
    d = dict()
    d.update(_PLOTTING_PARAMS)
    return d


def _base_plot(x, y, given_params) -> Axes:
    """Common function for creating plots.

    Returns:
        Axes object.

    """
    params = _get_plotting_params()
    params.update(given_params)

    linestyles = itertools.cycle(iter(params["linestyles"]))

    # Check if an axes object is passed in. Otherwise, create one.
    ax0 = params.get("ax", plt.figure().add_subplot(111))

    ax0.set_title(params["title"])
    if y.ndim > 1:
        for channel in y:
            ax0.plot(x, channel, linestyle=next(linestyles))
    else:
        ax0.plot(x, y)
    ax0.set_xlabel(params["xlabel"])
    ax0.set_ylabel(params["ylabel"])
    ax0.set_xscale(params["xscale"])
    ax0.set_yscale(params["yscale"])
    ax0.set_xlim(params["xlim"])
    ax0.set_ylim(params["ylim"])

    if params["labels"] is None and y.ndim > 1:
        params["labels"] = np.arange(y.shape[-2]) + 1
    if params["labels"] is not None:
        ax0.legend(labels=params["labels"])

    return ax0


__all__ = ["Signal"]



"""IEC 61260-1:2014 specifies performance requirements for analogue, sampled-data, and digital implementations of band-pass filters.

Functions:
    exact_center_frequency: Calculate exact center frequencies for given band indices
    lower_frequency: Calculate lower band-edge frequencies
    upper_frequency: Calculate upper band-edge frequencies
    index_of_frequency: Calculate the band index for a given frequency
    nominal_center_frequency: Calculate nominal center frequencies

Constants:
    - [NOMINAL_OCTAVE_CENTER_FREQUENCIES][acoustic_toolbox.standards.iec_61260_1_2014.NOMINAL_OCTAVE_CENTER_FREQUENCIES]: Nominal octave center frequencies
    - [NOMINAL_THIRD_OCTAVE_CENTER_FREQUENCIES][acoustic_toolbox.standards.iec_61260_1_2014.NOMINAL_THIRD_OCTAVE_CENTER_FREQUENCIES]: Nominal third-octave center frequencies
    - [REFERENCE_FREQUENCY][acoustic_toolbox.standards.iec_61260_1_2014.REFERENCE_FREQUENCY]: Reference frequency
    - [OCTAVE_FREQUENCY_RATIO][acoustic_toolbox.standards.iec_61260_1_2014.OCTAVE_FREQUENCY_RATIO]: Octave frequency ratio
"""

import acoustic_toolbox
import numpy as np

NOMINAL_OCTAVE_CENTER_FREQUENCIES = np.array(
    [31.5, 63.0, 125.0, 250.0, 500.0, 1000.0, 2000.0, 4000.0, 8000.0, 16000.0]
)
"""Nominal octave center frequencies."""

NOMINAL_THIRD_OCTAVE_CENTER_FREQUENCIES = np.array(
    [
        25.0,
        31.5,
        40.0,
        50.0,
        63.0,
        80.0,
        100.0,
        125.0,
        160.0,
        200.0,
        250.0,
        315.0,
        400.0,
        500.0,
        630.0,
        800.0,
        1000.0,
        1250.0,
        1600.0,
        2000.0,
        2500.0,
        3150.0,
        4000.0,
        5000.0,
        6300.0,
        8000.0,
        10000.0,
        12500.0,
        16000.0,
        20000.0,
    ]
)
"""Nominal third-octave center frequencies in the audio range."""

REFERENCE_FREQUENCY = 1000.0
"""Reference frequency."""

OCTAVE_FREQUENCY_RATIO = 10.0 ** (3.0 / 10.0)
"""Octave frequency ratio $G$.

See equation 1.
"""


def exact_center_frequency(
    x, fraction=1, ref=REFERENCE_FREQUENCY, G=OCTAVE_FREQUENCY_RATIO
) -> float | np.ndarray:
    r"""Center frequencies $f_m$ for band indices $x$. See equation 2 and 3.

    The center frequencies are given by
    $$
    f_m = f_r G^{x/b}
    $$

    In case the bandwidth designator $b$ is an even number, the center frequencies are given by
    $$
    f_m = f_r G^{(2x+1)/2b}
    $$

    See equation 2 and 3 of the standard.

    Args:
        x: Band index $x$.
        fraction: Bandwidth designator $b$. For example, for 1/3-octave filter $b=3$. Defaults to 1.
        ref: Reference center frequency $f_r$.
        G: Octave frequency ratio $G$.

    Returns:
        Center frequencies $f_m$ calculated according to equations 2 and 3.
    """
    fraction = np.asarray(fraction)
    uneven = (fraction % 2).astype("bool")
    return ref * G ** ((2.0 * x + 1.0) / (2.0 * fraction)) * np.logical_not(
        uneven
    ) + uneven * ref * G ** (x / fraction)


def lower_frequency(
    center: float | np.ndarray, fraction=1, G=OCTAVE_FREQUENCY_RATIO
) -> float | np.ndarray:
    """Lower band-edge frequencies. See equation 4.

    The lower band-edge frequencies are given by
    $$
    f_1 = f_m G^{-1/2b}
    $$

    See equation 4 of the standard.

    Args:
        center: Center frequencies $f_m$.
        fraction: Bandwidth designator $b$.
        G: Octave frequency ratio $G$.

    Returns:
        Lower band-edge frequencies $f_1$
    """
    return center * G ** (-1.0 / (2.0 * fraction))


def upper_frequency(
    center: float | np.ndarray, fraction=1, G=OCTAVE_FREQUENCY_RATIO
) -> float | np.ndarray:
    """Upper band-edge frequencies. See equation 5.

    The upper band-edge frequencies are given by
    $$
    f_2 = f_m G^{+1/2b}
    $$

    See equation 5 of the standard.

    Args:
        center: Center frequencies $f_m$.
        fraction: Bandwidth designator $b$.
        G: Octave frequency ratio $G$.

    Returns:
        Upper band-edge frequencies $f_2$.
    """
    return center * G ** (+1.0 / (2.0 * fraction))


def index_of_frequency(
    frequency, fraction=1, ref=REFERENCE_FREQUENCY, G=OCTAVE_FREQUENCY_RATIO
) -> int | np.ndarray:
    r"""Calculate the band index for a given frequency.

    The index of the center frequency is given by

    $$
    x = \text{round}\left(b \frac{\log(f/f_r)}{\log(G)}\right)
    $$

    See equation 6 of the standard.

    Note:
        This equation is not part of the standard.
        However, it follows from [`exact_center_frequency`](acoustic_toolbox.standards.iec_61260_1_2014.exact_center_frequency).

    Args:
        frequency: Frequencies $f$
        fraction: Bandwidth designator $b$.
        ref: Reference frequency.
        G: Octave frequency ratio $G$.

    Returns:
        Band indices $x$
    """
    fraction = np.asarray(fraction)
    uneven = (fraction % 2).astype("bool")
    return (
        np.round((2.0 * fraction * np.log(frequency / ref) / np.log(G) - 1.0))
        / 2.0
        * np.logical_not(uneven)
        + uneven
        * np.round(fraction * np.log(frequency / ref) / np.log(G)).astype("int16")
    ).astype("int16")


def _roundn(x, n):
    """Round a number to n significant digits.

    Args:
        x: Number to round.
        n: Number of significant digits.

    Returns:
        float: Rounded number.
    """
    return round(x, -int(np.floor(np.sign(x) * np.log10(abs(x)))) + n)


def _nominal_center_frequency(center, fraction):
    """Calculate nominal frequency according to standard.

    Args:
        center: Exact mid-frequency to be rounded.
        fraction: Bandwidth designator or fraction.

    Returns:
        float: Nominal center frequency according to standard sections E.1-E.3.

    See Also:
        exact_center_frequency: Function to calculate exact center frequencies

    Raises:
        NotImplementedError: If fraction > 24.
        ValueError: If fraction is invalid.
    """
    b = fraction
    x = center

    # Section E.1: 1/1-octaves
    if b == 1:
        n = index_of_frequency(x, b)
        if -6 <= n < 5:  # Correspond to indices when n=0 corresponds to 1000 Hz
            return acoustic_toolbox.standards.iec_61672_1_2013.NOMINAL_OCTAVE_CENTER_FREQUENCIES[
                n + 6
            ]
        elif n >= 5:
            return 2.0 * _nominal_center_frequency(
                exact_center_frequency(n - 1, b), b
            )  # WARNING: Unclear in standard!
        else:
            return (
                1.0
                / 2.0
                * _nominal_center_frequency(exact_center_frequency(n + 1, b), b)
            )  # WARNING: Unclear in standard!

    # Section E.2: 1/2-octaves
    elif b == 2:
        return _roundn(x, 2)

    # Section E.1: 1/3-octaves
    elif b == 3:
        n = index_of_frequency(x, b)

        if -20 <= n < 14:  # Correspond to indices when n=0 corresponds to 1000 Hz
            return acoustic_toolbox.standards.iec_61672_1_2013.NOMINAL_THIRD_OCTAVE_CENTER_FREQUENCIES[
                n + 20
            ]
        elif n >= 14:
            return 10.0 * _nominal_center_frequency(
                exact_center_frequency(n - 10, b), b
            )  # WARNING: Unclear in standard!
        else:
            return (
                1.0
                / 10.0
                * _nominal_center_frequency(exact_center_frequency(n + 10, b), b)
            )  # WARNING: Unclear in standard!

    # Section E3.3: 1/4 to 1/24-octaves, inclusive
    elif 4 <= b <= 24:
        msd = x // 10.0 ** np.floor(np.log10(x))
        if msd < 5:
            return _roundn(x, 2)  # E3.2
        else:
            return _roundn(x, 1)  # E3.3

    # Section E3.5: > 1/24-octaves
    elif b > 24:
        raise NotImplementedError("b > 24 is not implemented")
    else:
        raise ValueError("Wrong value for b")


nominal_center_frequency = np.vectorize(_nominal_center_frequency)
"""Vectorized function to calculate nominal center frequencies.

Args:
    center: Exact center frequency.
    fraction: Band designator or fraction.

Returns:
    float or ndarray: Nominal center frequencies calculated according to standard.
"""



"""The generator module provides signal generators.

The following functions calculate ``N`` samples and return an array containing the samples.

For indefinitely long iteration over the samples, consider using the output of these functions
in [`itertools.cycle`][itertools.cycle].

# Noise

Different types of noise are available. The following table lists the color
of noise and how the power and power density change per octave:

| Color  | Power | Power density |
|--------|:-----:|:-------------:|
| White  | +3 dB |    0 dB       |
| Pink   |  0 dB |   -3 dB       |
| Blue   | +6 dB |   +3 dB       |
| Brown  | -3 dB |   -6 dB       |
| Violet | +9 dB |   +6 dB       |

The colored noise is created by generating pseudo-random numbers using
[`np.random.randn`][numpy.random.randn] and then multiplying these with a curve typical for the color.
Afterwards, an inverse DFT is performed using [`np.fft.irfft`][numpy.fft.irfft].
Finally, the noise is normalized using [`acoustic_toolbox.signal.normalize`][acoustic_toolbox.signal.normalize].

## All colors

Functions:
    noise: Generate noise of a specified color.
    noise_generator: Generate `N` amount of unique samples and cycle over these samples.

## Per color

Functions:
    white: Generate white noise with constant power density and flat narrowband spectrum.
    pink: Generate pink noise with equal power in proportionally wide bands.
    blue: Generate blue noise with power increasing 6 dB per octave.
    brown: Generate brown noise with power decreasing -3 dB per octave.
    violet: Generate violet noise with power increasing +9 dB per octave.
    heaviside: Returns the value 0 for `x < 0`, 1 for `x > 0`, and 1/2 for `x = 0`.

See Also:
    For related functions, check [`scipy.signal`][scipy.signal].
"""

import itertools
from typing import Generator
import numpy as np

try:
    from pyfftw.interfaces.numpy_fft import (
        irfft,
    )  # Performs much better than numpy's fftpack
except ImportError:  # Use monkey-patching np.fft perhaps instead?
    from numpy.fft import irfft  # pylint: disable=ungrouped-imports

from .signal import normalize


def noise(
    N: int, color: str = "white", state: np.random.RandomState | None = None
) -> np.ndarray:
    """Noise generator.

    Args:
      N: Amount of samples.
      color: Color of noise.
      state: State of PRNG.

    Returns:
        Array of noise samples.
    """
    try:
        return _noise_generators[color](N, state)
    except KeyError:
        raise ValueError("Incorrect color.")


def white(N: int, state: np.random.RandomState | None = None) -> np.ndarray:
    """White noise.

    White noise has a constant power density. Its narrowband spectrum is therefore flat.
    The power in white noise will increase by a factor of two for each octave band,
    and therefore increases with 3 dB per octave.

    Args:
        N: Amount of samples.
        state: State of PRNG.

    Returns:
        Array of white noise samples.
    """
    state = np.random.RandomState() if state is None else state
    return state.randn(N)


def pink(N: int, state: np.random.RandomState | None = None) -> np.ndarray:
    """Pink noise.

    Pink noise has equal power in bands that are proportionally wide.
    Power density decreases with 3 dB per octave.

    Note:
        This method uses the filter with the following coefficients.
        $$
        B = [0.049922035, -0.095993537, 0.050612699, -0.004408786]
        $$
        $$
        A = [1, -2.494956002, 2.017265875, -0.522189400]
        $$

        The filter is applied using [`scipy.signal.lfilter`][scipy.signal.lfilter]:
        ```python
        from scipy.signal import lfilter
        b = np.array([0.049922035, -0.095993537, 0.050612699, -0.004408786])
        a = np.array([1, -2.494956002, 2.017265875, -0.522189400])
        return lfilter(b, a, np.random.randn(N))
        ```

        Another way would be using the FFT:
        ```python
        x = np.random.randn(N)
        X = rfft(x) / N
        S = np.sqrt(np.arange(len(X)) + 1.0)  # +1 to avoid divide by zero
        y = (irfft(X / S)).real
        ```

    Args:
        N: Amount of samples.
        state: State of PRNG.

    Returns:
        Array of pink noise samples.
    """
    # This method uses the filter with the following coefficients.
    # b = np.array([0.049922035, -0.095993537, 0.050612699, -0.004408786])
    # a = np.array([1, -2.494956002, 2.017265875, -0.522189400])
    # return lfilter(B, A, np.random.randn(N))
    # Another way would be using the FFT
    # x = np.random.randn(N)
    # X = rfft(x) / N
    state = np.random.RandomState() if state is None else state
    uneven = N % 2
    X = state.randn(N // 2 + 1 + uneven) + 1j * state.randn(N // 2 + 1 + uneven)
    S = np.sqrt(np.arange(len(X)) + 1.0)  # +1 to avoid divide by zero
    y = (irfft(X / S)).real
    if uneven:
        y = y[:-1]
    return normalize(y)


def blue(N: int, state: np.random.RandomState | None = None) -> np.ndarray:
    """Blue noise.

    Power increases with 6 dB per octave.
    Power density increases with 3 dB per octave.

    Args:
        N: Amount of samples.
        state: State of PRNG.

    Returns:
        Array of blue noise samples.
    """
    state = np.random.RandomState() if state is None else state
    uneven = N % 2
    X = state.randn(N // 2 + 1 + uneven) + 1j * state.randn(N // 2 + 1 + uneven)
    S = np.sqrt(np.arange(len(X)))  # Filter
    y = (irfft(X * S)).real
    if uneven:
        y = y[:-1]
    return normalize(y)


def brown(N: int, state: np.random.RandomState | None = None) -> np.ndarray:
    """Brown noise.

    Power decreases with -3 dB per octave.
    Power density decreases with 6 dB per octave.

    Args:
        N: Amount of samples.
        state: State of PRNG.

    Returns:
        Array of violet noise samples.
    """
    state = np.random.RandomState() if state is None else state
    uneven = N % 2
    X = state.randn(N // 2 + 1 + uneven) + 1j * state.randn(N // 2 + 1 + uneven)
    S = np.arange(len(X)) + 1  # Filter
    y = (irfft(X / S)).real
    if uneven:
        y = y[:-1]
    return normalize(y)


def violet(N: int, state: np.random.RandomState | None = None) -> np.ndarray:
    """Violet noise.

    Power increases with +9 dB per octave.
    Power density increases with +6 dB per octave.


    Args:
        N: Amount of samples.
        state: State of PRNG.

    Returns:
        Array of violet noise samples.
    """
    state = np.random.RandomState() if state is None else state
    uneven = N % 2
    X = state.randn(N // 2 + 1 + uneven) + 1j * state.randn(N // 2 + 1 + uneven)
    S = np.arange(len(X))  # Filter
    y = (irfft(X * S)).real
    if uneven:
        y = y[:-1]
    return normalize(y)


_noise_generators = {
    "white": white,
    "pink": pink,
    "blue": blue,
    "brown": brown,
    "violet": violet,
}


def noise_generator(
    N: int = 44100, color: str = "white", state: np.random.RandomState | None = None
) -> Generator[float, None, None]:
    """Noise generator.

    Generate `N` amount of unique samples and cycle over these samples.

    Args:
      N: Amount of unique samples to generate.
      color: Color of noise.
      state: State of PRNG.

    Returns:
        Generator of noise samples.
    """
    # yield from itertools.cycle(noise(N, color)) # Python 3.3
    for sample in itertools.cycle(noise(N, color, state)):
        yield sample


def heaviside(N: int) -> np.ndarray:
    """Heaviside.

    Returns the value 0 for `x < 0`, 1 for `x > 0`, and 1/2 for `x = 0`.

    Args:
      N: Amount of samples.

    Returns:
        Array of heaviside samples.
    """
    return 0.5 * (np.sign(N) + 1)


__all__ = [
    "noise",
    "white",
    "pink",
    "blue",
    "brown",
    "violet",
    "noise_generator",
    "heaviside",
]



import pytest
import numpy as np
from acoustic_toolbox.standards.iso_1996_2_2007 import Tonality
from acoustic_toolbox import Signal


def test_tonality():
    duration = 60.0
    fs = 10025.0
    samples = int(fs * duration)
    times = np.arange(samples) / fs

    signal = Signal(np.sin(2.0 * np.pi * 1000.0 * times), fs)

    tonality = Tonality(signal, signal.fs)

    # Test methods before analysis
    tonality.spectrum
    tonality.plot_spectrum()

    tonality.frequency_resolution
    tonality.effective_analysis_bandwidth

    # No values yet, cannot print overview.
    with pytest.raises(ValueError):
        print(tonality.overview())
    tonality.results_as_dataframe()

    assert len(list(tonality.noise_pauses)) == 0
    assert len(list(tonality.tones)) == 0
    assert len(list(tonality.critical_bands)) == 0

    # Perform analysis
    tonality.determine_noise_pauses().analyse()

    # Needs to be checked
    # assert len(list(tonality.noise_pauses)) == 1
    # assert len(list(tonality.tones)) == 1
    # assert len(list(tonality.critical_bands)) == 1

    tonality.critical_band_at(900.0)

    tonality.dominant_tone
    print(tonality.overview())
    tonality.results_as_dataframe()
    tonality.plot_results()


# Target = collections.namedtuple('Target', ['tonal_level',
#'masking_noise_level',
#'tonal_audibility',
#'adjustment'])
# TARGET_DATA = {
#'example1': Target(46.7, 37.3, 13.7, 6.0),
#'example2': Target(54.1, 45.2, 11.1, 6.0),
#'example3': Target(54.6, 45.5, 10.6, 6.0),
##'example4': Target(53.6, 45.5, 10.7, 6.0), # Time-varying test.
# }

# @pytest.fixture(params=TARGET_DATA.keys())
# def example(request):
# return request.param

# def test_verify_standard_examples(example):
# """Verify against the examples shown in the standard.


# .. note:: The FTP server is slow. Be patient.

# For the examples in the standard:
# - Number of spectra: 350
# - Signal duration: 2 minutes
# - Window: Hanning
# - Averaging: Linear
# - Effective analysis bandwidth: 4.39 Hz

# The sample frequency is 44.1 kHz. Dividing this sample frequency by
# results in 360 spectra and an effective analysis bandwidth of
# 4.5 Hz, which is very close to the example.

# """
# from ftplib import FTP
# import tempfile
# import os

# FOLDER = 'THP wave files/'
# FILES = {'calibration' : 'cal. 93,8 dB.wav',
#'example1': 'sam 03 eks 1.wav',
#'example2': 'sam 03 eks 2.wav',
#'example3': 'sam 03 eks 3.wav',
#'example4': 'sam 03 eks 4.wav',
# }
# CHANNEL = 0

# fs = 44100.0
# duration = 120.0
# samples = duration*fs
# spectra = 350
# bins = samples // spectra

## Obtain data from Delta
## https://noiselabdk.wordpress.com/demo-download/
##hostname = 'ftp.delta.dk'
##remote_data = 'Demo Files/Tone wave files.zip'
##username = 'nlpublic'
##password = 'noiselab'
##local_data = 'data.zip'

##with tempfile.TemporaryDirectory() as directory:
##ftp = FTP(hostname, username, password)
##with open(local_data, 'wb') as target_file:
##ftp.retrbinary('RETR %s' % remote_data, target_file.write)


# calibration_at = 93.8 # decibel
# signal = Signal.from_wav(os.path.join(FOLDER, FILES['calibration']))[CHANNEL]
# calibration_factor = 10.0**((calibration_at - signal.leq())/20.0)

# target = TARGET_DATA[example]
# filename = os.path.join(FOLDER, FILES[example])

# signal = Signal.from_wav(filename)[CHANNEL].pick(0.0, 60.0).weigh('A') * calibration_factor
# tonality = Tonality(signal, signal.fs, bins=bins)
# tonality.determine_noise_pauses().analyse()

# cb = tonality.dominant_tone.critical_band
# print(tonality.results())
# assert abs(tonality.dominant_tone.tone_level - target.tonal_level) < 1.0
# assert abs(cb.masking_noise_level - target.masking_noise_level) < 1.0
# assert abs(cb.tonal_audibility - target.tonal_audibility) < 1.0
# assert abs(cb.adjustment - target.adjustment) < 1.0



r"""The directivity module provides tools to work with directivity.

The following conventions are used within this module:

* The inclination angle $\theta$ has a range $[0, \pi]$.
* The azimuth angle $\phi$ has a range $[0 , 2 \pi]$.

Functions:
    cardioid: Generate a cardioid pattern.
    figure_eight: Generate a figure-of-eight pattern.
    spherical_harmonic: Calculate spherical harmonic of order `m` and degree `n`.
    spherical_to_cartesian: Convert spherical coordinates to cartesian coordinates.
    cartesian_to_spherical: Convert cartesian coordinates to spherical coordinates.

Classes:
    Directivity: Abstract class for directivity.
    Omni: Class for omni-directional directivity.
    Cardioid: Class for cardioid directivity.
    FigureEight: Class for figure-of-eight directivity.
    SphericalHarmonic: Class for spherical harmonic directivity.
    Custom: Class for custom directivity.
"""

import abc
import matplotlib.cm as cm
import matplotlib.pyplot as plt
from matplotlib.colors import Normalize
from matplotlib.figure import Figure
import numpy as np
from scipy.interpolate import interp2d as interpolate
from scipy.special import sph_harm  # pylint: disable=no-name-in-module


def cardioid(theta, a=1.0, k=1.0):
    r"""A cardioid pattern.

    Args:
      a: a
      k: k
      theta: angle $\theta$

    Returns:
      Cardioid pattern.
    """
    return np.abs(a + a * np.cos(k * theta))


def figure_eight(theta, phi=0.0):
    r"""A figure-of-eight pattern.

    Args:
      theta: angle $\theta$
      phi: angle $\phi$

    Returns:
      Figure-of-eight pattern.
    """
    del phi
    # return spherical_harmonic(theta, phi, m=0, n=1)
    return np.abs(np.cos(theta))


def spherical_harmonic(theta, phi, m=0, n=0):
    """Spherical harmonic of order `m` and degree `n`.

    Note:
      The degree `n` is often denoted `l`.

    See Also:
      [`scipy.special.sph_harm`][scipy.special.sph_harm]
    """
    # TODO: `scipy.special.sph_harm` is deprecated since v1.15. Should use `sph_harm_y` instead.
    # https://docs.scipy.org/doc/scipy/reference/generated/scipy.special.sph_harm.html
    # TODO: `scipy` expects `m` and `n` to be integers. Will need to adapt typehints
    return sph_harm(m, n, phi, theta).real


def spherical_to_cartesian(r, theta, phi):
    r"""Convert spherical coordinates to cartesian coordinates.

    Args:
      r: norm
      theta: angle $\theta$
      phi: angle $\phi$

    Returns:
       x: $x = r \sin{\theta}\cos{\phi}$
       y: $y = r \sin{\theta}\sin{\phi}$
       z: $z = r \cos{\theta}$
    """
    r = np.asanyarray(r)
    theta = np.asanyarray(theta)
    phi = np.asanyarray(phi)
    return (
        r * np.sin(theta) * np.cos(phi),
        r * np.sin(theta) * np.sin(phi),
        r * np.cos(theta),
    )


def cartesian_to_spherical(x, y, z):
    r"""Convert cartesian coordinates to spherical coordinates.

    Args:
      x: x
      y: y
      z: z

    Returns:
      r: $r = \sqrt{\left( x^2 + y^2 + z^2 \right)}$
      theta: $\theta = \arccos{\frac{z}{r}}$
      phi: $\phi = \arccos{\frac{y}{x}}$
    """
    x = np.asanyarray(x)
    y = np.asanyarray(y)
    z = np.asanyarray(z)
    r = np.linalg.norm(np.vstack((x, y, z)), axis=0)
    return r, np.arccos(z / r), np.arctan(y / x)


class Directivity:
    """Abstract directivity class.

    This class defines several methods to be implemented by subclasses.
    """

    def __init__(self, rotation=None):
        """Rotation of the directivity pattern.

        Args:
          rotation: Rotation of the directivity pattern.
        """
        self.rotation = (
            rotation if rotation else np.array([1.0, 0.0, 0.0])
        )  # X, Y, Z rotation

    @abc.abstractmethod
    def _directivity(self, theta, phi):
        r"""This function should return the directivity as function of $\theta$ and $\phi$."""

    def _undo_rotation(self, theta, phi):
        """Undo rotation."""

    def using_spherical(self, theta, phi, r=None, include_rotation: bool = True):
        r"""Return the directivity for given spherical coordinates.

        Args:
          theta: angle $\theta$
          phi: angle $\phi$
          r: norm (optional)
          include_rotation: Apply the rotation to the directivity.

        Returns:
          Directivity.

        Todo:
          Correct for rotation!!!!
        """
        # TODO: Correct for rotation!!!! use 'r' if needed in rotation logic
        return self._directivity(theta, phi)

    def using_cartesian(self, x, y, z, include_rotation: bool = True):
        """Return the directivity for given cartesian coordinates.

        Args:
          x: x
          y: y
          z: z
          include_rotation: Apply the rotation to the directivity.

        Returns:
          Directivity.

        Todo:
          Correct for rotation!!!!
        """
        r, theta, phi = cartesian_to_spherical(x, y, z)
        return self.using_spherical(theta, phi, r, include_rotation)

    def plot(
        self, filename: str | None = None, include_rotation: bool = True
    ) -> Figure:
        """Directivity plot. Plot to ``filename`` when given.

        Args:
          filename: Filename
          include_rotation: Apply the rotation to the directivity.
            By default the rotation is applied in this figure.

        Returns:
          Figure.

        Todo:
          filename
        """
        raise NotImplementedError(
            "`Directivity.plot` needs to be re-implemented, due to underlying issues from python-acoustics."
        )
        # I'm not sure how the example notebook ever ran before...
        # The method needs to be re-implemented to match the `plot` function.
        del filename
        return plot(self, include_rotation)


class Omni(Directivity):
    """Class to work with omni-directional directivity."""

    def _directivity(self, theta, phi):
        """Directivity."""
        return np.ones_like(theta)


class Cardioid(Directivity):
    """Cardioid directivity."""

    def _directivity(self, theta, phi):
        """Directivity."""
        return cardioid(theta)


class FigureEight(Directivity):
    """Directivity of a figure of eight."""

    def _directivity(self, theta, phi):
        """Directivity."""
        return figure_eight(theta, phi)


class SphericalHarmonic(Directivity):
    """Directivity of a spherical harmonic of degree `n` and order `m`."""

    def __init__(self, rotation=None, m=None, n=None):
        """Constructor."""
        super().__init__(rotation=rotation)
        self.m = m
        """Order `m`."""
        self.n = n
        """Degree `n`."""

    def _directivity(self, theta, phi):
        """Directivity."""
        return spherical_harmonic(theta, phi, self.m, self.n)


class Custom(Directivity):
    """A class to work with directivity."""

    def __init__(self, theta=None, phi=None, r=None):
        """Constructor."""
        self.theta = theta
        """Latitude. 1-D array."""

        self.phi = phi
        """Longitude. 1-D array."""
        self.r = r
        """Magnitude or radius. 2-D array."""

    def _directivity(self, theta, phi):
        """Custom directivity.

        Interpolate the directivity given longitude and latitude vectors.
        """
        f = interpolate(self.theta, self.phi, self.r)

        return f(theta, phi)


def plot(d: Directivity, sphere: bool = False) -> Figure:
    """Plot directivity `d`.

    Args:
      d: Directivity
      sphere: Plot on a sphere.

    Returns:
      Figure
    """
    theta, phi = np.meshgrid(
        np.linspace(0.0, np.pi, 50), np.linspace(0.0, +2.0 * np.pi, 50)
    )

    # Directivity strength. Real-valued. Can be positive and negative.
    dr = d.using_spherical(theta, phi)

    if sphere:
        x, y, z = spherical_to_cartesian(1.0, theta, phi)

    else:
        x, y, z = spherical_to_cartesian(np.abs(dr), theta, phi)
    # R, theta, phi = cartesian_to_spherical(x, y, z)

    fig = plt.figure()
    ax = fig.add_subplot(111, projection="3d")
    # p = ax.plot_surface(x, y, z, cmap=plt.cm.jet, rstride=1, cstride=1, linewidth=0)

    norm = Normalize()
    norm.autoscale(dr)
    colors = cm.jet(norm(dr))
    m = cm.ScalarMappable(cmap=cm.jet, norm=norm)
    m.set_array(dr)
    ax.plot_surface(x, y, z, facecolors=colors, rstride=1, cstride=1, linewidth=0)
    plt.colorbar(m, ax=ax)

    ax.set_xlabel("$x$")
    ax.set_ylabel("$y$")
    ax.set_zlabel("$z$")
    return fig



"""Tests for :func:`acoustic_toolbox.signal`"""

from acoustic_toolbox.signal import convolve as convolveLTV
from scipy.signal import convolve as convolveLTI
import numpy as np

from acoustic_toolbox.signal import (
    EqualBand,
    OctaveBand,
    decibel_to_neper,
    neper_to_decibel,
    ir2fr,
    zero_crossings,
    integrate_bands,
    bandpass_octaves,
    bandpass_third_octaves,
    bandpass_fractional_octaves,
    bandpass_frequencies,
    amplitude_envelope,
    rms,
    ms,
)
from numpy.testing import assert_array_almost_equal, assert_array_equal

import pytest


@pytest.mark.parametrize(
    "u,h",
    [
        (
            np.array([1, 2, 3, 4, 3, 2, 1], dtype="float"),
            np.array([1, 2, 3, 4], dtype="float"),
        ),
        (
            np.array([1, 2, 3, 4, 3, 2, 1, 1], dtype="float"),
            np.array([1, 2, 3, 4, 5], dtype="float"),
        ),
    ],
)
def test_convolve_lti(u, h):
    """Test whether :func:`acoustic_toolbox.signal.convolve` behaves properly when
    performing a convolution with a time-invariant system.
    """
    H = np.tile(h, (len(u), 1)).T

    np.testing.assert_array_almost_equal(convolveLTV(u, H), convolveLTI(u, h))
    np.testing.assert_array_almost_equal(
        convolveLTV(u, H, mode="full"), convolveLTI(u, h, mode="full")
    )
    np.testing.assert_array_almost_equal(
        convolveLTV(u, H, mode="valid"), convolveLTI(u, h, mode="valid")
    )
    # This function and test needs some investigation. Disabling for now.
    # np.testing.assert_array_almost_equal(convolveLTV(u,H,mode='same'), convolveLTI(u,h,mode='same'))


def test_convolve_ltv():
    """Test whether :func:`acoustic_toolbox.signal.convolve` behaves properly when
    performing a convolution with a time-variant system.
    """
    """Input signal"""
    u = np.array([1, 1, 1])
    """Impulse responses where each column represents an impulse response."""
    C = np.array([[1, 0, 0], [2, 1, 1]])
    """The result calculated manually."""
    y_manual = np.array([1, 2, 1, 1])

    y_ltv = convolveLTV(u, C)
    np.testing.assert_array_equal(y_ltv, y_manual)


def test_decibel_to_neper():
    assert decibel_to_neper(1.0) == 0.11512925464970229


def test_neper_to_decibel():
    assert neper_to_decibel(1.0) == 8.685889638065035


def test_ir2fr():
    """Test whether the frequency vector is correct."""
    t = 1.0
    fs = 100.0
    f = 20.0
    ts = np.arange(0, t, 1.0 / fs)

    A = 5.0

    x = A * np.sin(2.0 * np.pi * f * ts)

    fv, fr = ir2fr(x, fs)

    assert_array_almost_equal(fv[np.abs(fr).argmax()], f)

    assert_array_almost_equal(np.abs(fr).max(), A)


class TestEqualBand:  # (unittest.TestCase):
    """Test :class:`acoustic_toolbox.signal.EqualBand`."""

    def test_construction_1(self):
        """Using center."""
        x = np.arange(10.0, 20.0, 2.0)
        b = EqualBand(x)
        assert_array_equal(b.center, x)

    def test_construction_2(self):
        """Using fstart, fstop and fbands"""
        x = np.arange(10.0, 20.0, 2.0)
        fstart = x[0]
        fstop = x[-1]
        nbands = len(x)
        b = EqualBand(fstart=fstart, fstop=fstop, nbands=nbands)
        assert_array_equal(b.center, x)

    def test_construction_3(self):
        """Using fstart, fstop and bandwidth"""
        x = np.arange(10.0, 20.0, 2.0)
        fstart = x[0]
        fstop = x[-1]
        bandwidth = np.diff(x)[0]
        b = EqualBand(fstart=fstart, fstop=fstop, bandwidth=bandwidth)
        assert_array_equal(b.center, x)

    def test_construction_4(self):
        # Using fstart, bandwidth and bands
        x = np.arange(10.0, 20.0, 2.0)
        fstart = x[0]
        bandwidth = np.diff(x)[0]
        nbands = len(x)
        b = EqualBand(fstart=fstart, nbands=nbands, bandwidth=bandwidth)
        assert_array_equal(b.center, x)

    def test_construction_5(self):
        # Using fstop, bandwidth and bands
        x = np.arange(10.0, 20.0, 2.0)
        fstop = x[-1]
        bandwidth = np.diff(x)[0]
        nbands = len(x)
        b = EqualBand(fstop=fstop, nbands=nbands, bandwidth=bandwidth)
        assert_array_equal(b.center, x)

    def test_selection(self):
        eb = EqualBand(fstart=0.0, fstop=10.0, nbands=100)
        assert type(eb[3] == type(eb))
        assert type(eb[3:10] == type(eb))


class Test_integrate_bands:
    """Test :func:`acoustic_toolbox.signal.test_integrate_bands`."""

    def test_narrowband_to_octave(self):
        nb = EqualBand(np.arange(100, 900, 200.0))
        x = np.ones(len(nb))
        ob = OctaveBand(([125.0, 250, 500.0]))
        y = integrate_bands(x, nb, ob)
        assert_array_equal(y, np.array([1, 1, 2]))


def test_zero_crossings():
    duration = 2.0
    fs = 44100.0
    f = 1000.0
    samples = int(duration * fs)
    t = np.arange(samples) / fs
    x = np.sin(2.0 * np.pi * f * t)

    z = zero_crossings(x)

    # Amount of zero crossings.
    assert len(z) == f * duration * 2

    # Position of zero crossings.
    y = np.linspace(0, samples, len(z), endpoint=False).astype(int)
    assert (np.abs(z - y) <= 1).all()


def test_ms():
    duration = 2.0
    fs = 8000.0
    f = 1000.0
    samples = int(duration * fs)
    t = np.arange(samples) / fs
    x = np.sin(2.0 * np.pi * f * t)

    assert np.abs(ms(x) - 0.5) < 1e-9

    x *= 4.0

    assert np.abs(ms(x) - 8.0) < 1e-9


def test_rms():
    duration = 2.0
    fs = 8000.0
    f = 1000.0
    samples = int(duration * fs)
    t = np.arange(samples) / fs
    x = np.sin(2.0 * np.pi * f * t)

    assert np.abs(rms(x) - np.sqrt(0.5)) < 1e-9

    x *= 4.0

    assert np.abs(rms(x) - np.sqrt(8.0)) < 1e-9


@pytest.fixture(params=[4000.0, 20000.0, 44100.0])
def fs(request):
    return request.param


@pytest.fixture(params=[1.0, 2.0, 3.0])
def amplitude(request):
    return request.param


@pytest.fixture(params=[100.0, 200.0, 300.0])
def frequency(request):
    return request.param


def test_amplitude_envelope(amplitude, frequency, fs):
    """Test amplitude envelope."""
    duration = 5.0
    samples = int(fs * duration)
    t = np.arange(samples) / fs

    signal = amplitude * np.sin(2.0 * np.pi * frequency * t)

    out = amplitude_envelope(signal, fs)
    # Rounding is necessary. We take the first element because occasionally
    # there is also a zero.
    amplitude_determined = np.unique(np.round(out), 6)[0]

    assert amplitude == amplitude_determined


# def test_instantaneous_frequency(amplitude, frequency, fs):

# duration = 5.0
# samples = int(fs*duration)
# t = np.arange(samples) / fs

# signal = amplitude * np.sin(2.0*np.pi*frequency*t)

# out = instantaneous_frequency(signal, fs)
## Rounding is necessary. We take the first element because occasionally
## there is also a zero.
# frequency_determined = np.unique(np.round(out), 0)

# assert( frequency == frequency_determined )


@pytest.mark.parametrize("channels", [1, 2, 5])
def test_bandpass(channels):
    fs = 88200
    duration = 2
    samples = duration * fs

    signal = np.random.randn(channels, samples)

    # We check whether it computes, and whether channels is in right dimension
    result = bandpass_octaves(signal, fs, order=8, purge=False)
    assert result[1].shape[-2] == channels

    result = bandpass_third_octaves(signal, fs, order=8, purge=False)
    assert result[1].shape[-2] == channels

    # We need to define frequencies explicitly
    with pytest.raises(TypeError):
        bandpass_fractional_octaves(signal, fs)

    frequencies = OctaveBand(fstart=100.0, fstop=2000.0, fraction=12)
    result = bandpass_fractional_octaves(signal, fs, frequencies, order=8, purge=False)
    assert result[1].shape[-2] == channels

    frequencies = EqualBand(center=[100.0, 200.0, 300.0], bandwidth=20.0)
    result = bandpass_frequencies(signal, fs, frequencies)
    assert result[1].shape[-2] == channels


@pytest.fixture(params=[1, 3, 6, 12, 24])
def fraction(request):
    return request.param


@pytest.fixture
def ob(fraction):
    return OctaveBand(fstart=10.0, fstop=1000, fraction=fraction)


class TestOctaveBand:
    def test_unique(self, ob):
        """Test whether we don't have duplicate values."""
        assert len(ob.center) == len(np.unique(ob.center))
        assert len(ob.lower) == len(np.unique(ob.lower))
        assert len(ob.upper) == len(np.unique(ob.upper))
        assert len(ob.nominal) == len(np.unique(ob.nominal))
        assert len(ob.bandwidth) == len(np.unique(ob.bandwidth))



"""The room acoustics module contains several functions to calculate the reverberation time in spaces."""

import numpy as np
from numpy.typing import NDArray

from scipy.io import wavfile
from scipy import stats

from acoustic_toolbox.utils import _is_1d
from acoustic_toolbox.signal import bandpass
from acoustic_toolbox.bands import (
    _check_band_type,
    octave_low,
    octave_high,
    third_low,
    third_high,
)

SOUNDSPEED = 343.0


def mean_alpha(alphas, surfaces):
    r"""Calculate mean of absorption coefficients.

    Args:
      alphas: Absorption coefficients $\alpha$.
      surfaces: Surfaces $S$.

    Returns:
      Mean absorption coefficient $\alpha_{m}$.
    """
    axis = None if np.ndim(alphas) == 0 else 0
    return np.average(alphas, axis=axis, weights=surfaces)


def nrc(alphas):
    r"""Calculate Noise Reduction Coefficient (NRC) from four absorption coefficient values (250, 500, 1000 and 2000 Hz).

    Args:
      alphas: Absorption coefficients $\alpha$.

    Returns:
      NRC
    """
    alpha_axis = alphas.ndim - 1
    return np.mean(alphas, axis=alpha_axis)


def t60_sabine(
    surfaces: NDArray[np.float64],
    alpha: NDArray[np.float64],
    volume: float,
    c: float = SOUNDSPEED,
):
    r"""Reverberation time according to Sabine.

    Sabine's formula for the reverberation time is:
      $$
      T_{60} = \frac{24 \ln(10)}{c} \frac{V}{S\alpha}
      $$

    Args:
      surfaces: Surface of the room $S$. NumPy array that contains different surfaces.
      alpha: Absorption coefficient of the room $\alpha$.
        Contains absorption coefficients of ``surfaces``.
        It could be one value or some values in different bands (1D and 2D
        array, respectively).
      volume: Volume of the room $V$.
      c: Speed of sound $c$.

    Returns:
      Reverberation time $T_{60}$


    """
    mean_alpha_ = np.average(alpha, axis=0, weights=surfaces)
    S = np.sum(surfaces, axis=0)
    A = S * mean_alpha_
    t60 = 4.0 * np.log(10.0**6.0) * volume / (c * A)
    return t60


def t60_eyring(
    surfaces: NDArray[np.float64],
    alpha: NDArray[np.float64],
    volume: float,
    c: float = SOUNDSPEED,
):
    r"""Reverberation time according to Eyring.

      Eyring's formula for the reverberation time is:
      $$
      T_{60} = \frac{24 \ln{10} V}{c \left( 4 mV - S \ln{\left( 1 - \alpha \right)} \right)}
      $$

    Args:
      surfaces: Surfaces $S$.
      alpha: Mean absorption coefficient $\alpha$ or by frequency bands
      volume: Volume of the room $V$.
      c: Speed of sound $c$.

    Returns:
      Reverberation time $T_{60}$

    """
    mean_alpha_ = np.average(alpha, axis=0, weights=surfaces)
    S = np.sum(surfaces, axis=0)
    A = -S * np.log(1 - mean_alpha_)
    t60 = 4.0 * np.log(10.0**6.0) * volume / (c * A)
    return t60


def t60_millington(
    surfaces: NDArray[np.float64],
    alpha: NDArray[np.float64],
    volume: float,
    c: float = SOUNDSPEED,
):
    r"""Reverberation time according to Millington.

    Args:
      surfaces: Surfaces $S$.
      alpha: Mean absorption coefficient $\alpha$ or by frequency bands
      volume: Volume of the room $V$.
      c: Speed of sound $c$.

    Returns:
      Reverberation time $T_{60}$

    """
    mean_alpha_ = np.average(alpha, axis=0, weights=surfaces)
    A = -np.sum(surfaces[:, np.newaxis] * np.log(1.0 - mean_alpha_), axis=0)
    t60 = 4.0 * np.log(10.0**6.0) * volume / (c * A)
    return t60


def t60_fitzroy(
    surfaces: NDArray[np.float64],
    alpha: NDArray[np.float64],
    volume: float,
    c: float = SOUNDSPEED,
):
    r"""Reverberation time according to Fitzroy.

    Args:
      surfaces: Surfaces $S$.
      alpha: Mean absorption coefficient $\alpha$ or by frequency bands
      volume: Volume of the room $V$.
      c: Speed of sound $c$.

    Returns:
      Reverberation time $T_{60}$

    """
    Sx = np.sum(surfaces[0:2])
    Sy = np.sum(surfaces[2:4])
    Sz = np.sum(surfaces[4:6])
    St = np.sum(surfaces)
    alpha = _is_1d(alpha)
    a_x = np.average(alpha[:, 0:2], weights=surfaces[0:2], axis=1)
    a_y = np.average(alpha[:, 2:4], weights=surfaces[2:4], axis=1)
    a_z = np.average(alpha[:, 4:6], weights=surfaces[4:6], axis=1)
    factor = -(Sx / np.log(1.0 - a_x) + Sy / np.log(1.0 - a_y) + Sz / np.log(1 - a_z))
    t60 = 4.0 * np.log(10.0**6.0) * volume * factor / (c * St**2.0)
    return t60


def t60_arau(
    Sx: float,
    Sy: float,
    Sz: float,
    alpha: tuple[float, float, float],
    volume: float,
    c: float = SOUNDSPEED,
):
    r"""Reverberation time according to Arau.

    For more details, please see
    [http://www.arauacustica.com/files/publicaciones/pdf_esp_7.pdf](http://www.arauacustica.com/files/publicaciones/pdf_esp_7.pdf)

    Args:
      Sx: Total surface perpendicular to x-axis (yz-plane) $S_{x}$.
      Sy: Total surface perpendicular to y-axis (xz-plane) $S_{y}$.
      Sz: Total surface perpendicular to z-axis (xy-plane) $S_{z}$.
      alpha: Absorption coefficients $\mathbf{\alpha} = \left[ \alpha_x, \alpha_y, \alpha_z \right]$
      volume: Volume of the room $V$.
      c: Speed of sound $c$.

    Returns:
      Reverberation time $T_{60}$



    """
    a_x = -np.log(1 - alpha[0])
    a_y = -np.log(1 - alpha[1])
    a_z = -np.log(1 - alpha[2])
    St = np.sum(np.array([Sx, Sy, Sz]))
    A = St * a_x ** (Sx / St) * a_y ** (Sy / St) * a_z ** (Sz / St)
    t60 = 4.0 * np.log(10.0**6.0) * volume / (c * A)
    return t60


def t60_impulse(
    file_name: str, bands: NDArray[np.float64], rt: str = "t30"
) -> NDArray[np.float64]:
    """Reverberation time from a WAV impulse response.

    Args:
      file_name: name of the WAV file containing the impulse response.
      bands: Octave or third bands as NumPy array.
      rt: Reverberation time estimator. It accepts `'t30'`, `'t20'`, `'t10'` and `'edt'`.

    Returns:
      Reverberation time $T_{60}$
    """
    fs, raw_signal = wavfile.read(file_name)
    band_type = _check_band_type(bands)

    if band_type == "octave":
        low = octave_low(bands[0], bands[-1])
        high = octave_high(bands[0], bands[-1])
    elif band_type == "third":
        low = third_low(bands[0], bands[-1])
        high = third_high(bands[0], bands[-1])

    rt = rt.lower()
    if rt == "t30":
        init = -5.0
        end = -35.0
        factor = 2.0
    elif rt == "t20":
        init = -5.0
        end = -25.0
        factor = 3.0
    elif rt == "t10":
        init = -5.0
        end = -15.0
        factor = 6.0
    elif rt == "edt":
        init = 0.0
        end = -10.0
        factor = 6.0

    t60 = np.zeros(bands.size)

    for band in range(bands.size):
        # Filtering signal
        filtered_signal = bandpass(raw_signal, low[band], high[band], fs, order=8)
        abs_signal = np.abs(filtered_signal) / np.max(np.abs(filtered_signal))

        # Schroeder integration
        sch = np.cumsum(abs_signal[::-1] ** 2)[::-1]
        sch_db = 10.0 * np.log10(sch / np.max(sch))

        # Linear regression
        sch_init = sch_db[np.abs(sch_db - init).argmin()]
        sch_end = sch_db[np.abs(sch_db - end).argmin()]
        init_sample = np.where(sch_db == sch_init)[0][0]
        end_sample = np.where(sch_db == sch_end)[0][0]
        x = np.arange(init_sample, end_sample + 1) / fs
        y = sch_db[init_sample : end_sample + 1]
        slope, intercept = stats.linregress(x, y)[0:2]

        # Reverberation time (T30, T20, T10 or EDT)
        db_regress_init = (init - intercept) / slope
        db_regress_end = (end - intercept) / slope
        t60[band] = factor * (db_regress_end - db_regress_init)
    return t60


def clarity(
    time: float, signal, fs: int, bands: NDArray[np.float64]
) -> NDArray[np.float64]:
    """Clarity $C_i$ determined from an impulse response.

    Args:
      time: Time in miliseconds (e.g.: 50, 80).
      signal: Impulse response.
      fs: Sample frequency.
      bands: Bands of calculation (optional). Only support standard octave and third-octave bands.

    Returns:
      Clarity $C_i$
    """
    # TODO: Not tested
    band_type = _check_band_type(bands)

    if band_type == "octave":
        low = octave_low(bands[0], bands[-1])
        high = octave_high(bands[0], bands[-1])
    elif band_type == "third":
        low = third_low(bands[0], bands[-1])
        high = third_high(bands[0], bands[-1])

    c = np.zeros(bands.size)
    for band in range(bands.size):
        filtered_signal = bandpass(signal, low[band], high[band], fs, order=8)
        h2 = filtered_signal**2.0
        t = int((time / 1000.0) * fs + 1)
        c[band] = 10.0 * np.log10((np.sum(h2[:t]) / np.sum(h2[t:])))
    return c


def c50_from_file(file_name: str, bands: NDArray[np.float64]) -> NDArray[np.float64]:
    """Clarity for 50 miliseconds $C_{50}$ from a file.

    Args:
      file_name: File name (only WAV is supported).
      bands: Bands of calculation (optional). Only support standard octave and third-octave bands.

    Returns:
      Clarity $C_{50}$
    """
    fs, signal = wavfile.read(file_name)
    return clarity(50.0, signal, fs, bands)


def c80_from_file(file_name: str, bands: np.ndarray):
    """Clarity for 80 miliseconds $C_{80}$ from a file.

    Args:
      file_name: File name (only WAV is supported).
      bands: Bands of calculation (optional). Only support standard octave and third-octave bands.

    Returns:
      Clarity $C_{80}$
    """
    fs, signal = wavfile.read(file_name)
    return clarity(80.0, signal, fs, bands)



"""Acoustic Toolbox
================

The acoustic_toolbox module.

"""

import acoustic_toolbox.ambisonics
import acoustic_toolbox.atmosphere
import acoustic_toolbox.bands
import acoustic_toolbox.building
import acoustic_toolbox.cepstrum
import acoustic_toolbox.criterion
import acoustic_toolbox.decibel
import acoustic_toolbox.descriptors
import acoustic_toolbox.directivity
import acoustic_toolbox.doppler
import acoustic_toolbox.generator
import acoustic_toolbox.imaging
import acoustic_toolbox.octave
import acoustic_toolbox.power
import acoustic_toolbox.quantity
import acoustic_toolbox.reflection
import acoustic_toolbox.room
import acoustic_toolbox.signal

# import acoustic_toolbox.utils
import acoustic_toolbox.weighting

from acoustic_toolbox._signal import Signal

__version__ = "0.2.0"
__all__ = ["Signal"]



"""This module implements ISO 1683:2015 which specifies preferred reference values for
levels used in acoustics and vibration.
"""

REFERENCE_VALUES = {
    "gas": {
        "pressure": 2e-5,  # Pa
        "exposure": (2e-5) ** 2,  # Pa^2 s
        "power": 1e-12,  # W
        "energy": 1e-12,  # J
        "intensity": 1e-12,  # W/m^2
    },
    "liquid": {
        "pressure": 1e-5,  # Pa
        "exposure": 1e-5,  # Pa^2 s
        "power": 1e-12,  # W
        "energy": 1e-12,  # J
        "intensity": 1e-12,  # W/m^2
        "particle_displacement": 1e-12,  # m
        "particle_velocity": 1e-9,  # m/s
        "particle_acceleration": 1e-6,  # m/s^2
        "distance": 1,  # m
    },
    "vibration": {
        "displacement": 1e-12,  # m
        "velocity": 1e-9,  # m/s
        "acceleration": 1e-6,  # m/s^2
        "force": 1e-6,  # N}
    },
}
